#include <unistd.h>
#include <qdom.h>
#include <qfile.h>
#include <qmessagebox.h>
#include <qlistview.h>
#include <qlineedit.h>
#include <qlabel.h>
#include <qthread.h>
#include <qeventloop.h>
#include <qtextedit.h>
#include <qcheckbox.h>
#include <qsettings.h>
#include "term.h"
#include "dialerext.h"
#include "dialerdialog.h"
#include "design/entryeditui.h"
#include "moc_dialdirectory.h"

DialDirectory::DialDirectory( int fd0, QWidget *parent ):
	fd(fd0), DialDirectoryUI( parent, "DialDirectoryDialog", true ),
	dialListModified(false), connectIdx(-1) {
  QListViewItem *it;
  int w;

  setFixedSize( size() );
  infoL->setText("");
  loadList();
  show();

  w = dirLV->visibleWidth();
  dirLV->setColumnWidth( 0, (int)(w*.4) );
  dirLV->setColumnWidth( 1, (int)(w*.2) );
  dirLV->setColumnWidth( 2, (int)(w*.15) );
  dirLV->setColumnWidth( 3, (int)(w*.25) );

  it = dirLV->firstChild();
  dirLV->setCurrentItem( it );
  dirLV->setSelected( it, true );
}

DialDirectory::~DialDirectory() {
  if ( dialListModified ) saveList();
}

DialEntry *DialDirectory::connectedTo() {
  if ( connectIdx == -1 ) return 0;
  return &dialList[connectIdx];
}

int monthIdx( QString month ) {
  int i;

  for ( i = 1; i <= 12; i++ )
  	if ( month == QDate::shortMonthName( i ) ) return i;

  return 1;
}

void DialDirectory::loadList() {
  QDomDocument doc("DialingList");
  int i;
  QString fname;

  dirLV->clear();

  fname = ConfigManager::getFilename( "dialdir.xml" );
  QFile f( fname );
  if ( !f.open( IO_ReadOnly ) ) return;

  doc.setContent( &f );

  QDomElement e, docElem = doc.documentElement();
  QDomNodeList nl = docElem.elementsByTagName( "BBSRecord" );

  QListViewItem *it = 0;
  DialEntry entry;
  QString calls, lastLogin, month;
  int day, year;
  for ( i = 0; i < nl.count(); i++ ) {
  	e = nl.item(i).toElement();

	entry.name = e.attribute("name");
	entry.phone = e.attribute("phone");

	entry.iemsiLogin = (e.attribute("iemsiLogin").toInt() != 0);
	entry.iemsiLoginName = e.attribute("iemsiLoginName");
	entry.iemsiPassword = e.attribute("iemsiPassword");

	calls = e.attribute("calls");
	entry.calls = calls.toInt();

	if ( entry.calls ) {
		lastLogin = e.attribute("lastLogin");
		day = lastLogin.section( '-', 0, 0 ).toInt();
		month = lastLogin.section( '-', 1, 1 );
		year = lastLogin.section( '-', 2, 2 ).toInt();
		entry.lastLogin = QDate( year, monthIdx(month), day );
	}
	else lastLogin = "";

	entry.dlSuffix = e.attribute("dlSuffix");

	QDomNodeList nl2 = e.elementsByTagName( "Notes" );
	if ( nl2.count() ) {
		QDomElement notes;
	
		notes = nl2.item(0).toElement();
		entry.notes = notes.text();
	}
	else entry.notes = "";

	dialList.push_back( entry );
	
  	it = new DirLVItem( i, dirLV, it,
		entry.name, entry.phone, calls, lastLogin );
  }

  updateNotesDisplay( dirLV->currentItem() );
}

void DialDirectory::saveList() {
  QDomDocument doc("DialingList");
  int i;
  QString fname;

  fname = ConfigManager::getFilename( "dialdir.xml" );
  QFile f( fname );
  if ( !f.open( IO_WriteOnly | IO_Truncate ) ) {
  	QMessageBox::warning( this, "Dialing Directory",
		"Unable to open dialdir.xml for writing",
		QMessageBox::Ok, 0 );
  	return;
  }

  QDomElement docElem, e;
  docElem = doc.createElement("DialingList");
  doc.appendChild( docElem );

  DialEntry entry;
  for ( i = 0; i < dialList.size(); i++ ) {
  	entry = dialList[i];

  	e = doc.createElement("BBSRecord");
	e.setAttribute( "name", entry.name );
	e.setAttribute( "phone", entry.phone );

	e.setAttribute( "iemsiLogin", entry.iemsiLogin? 1: 0 );
	e.setAttribute( "iemsiLoginName", entry.iemsiLoginName );
	e.setAttribute( "iemsiPassword", entry.iemsiPassword );
	
	e.setAttribute( "calls", entry.calls );
	if ( entry.calls )
		e.setAttribute( "lastLogin", entry.lastLogin.toString("dd-MMM-yyyy") );
	if ( !entry.dlSuffix.isNull() )
		e.setAttribute( "dlSuffix", entry.dlSuffix );
	if ( !entry.notes.isEmpty() ) {
		QDomElement notes;
		QDomText txtNode;
	
		notes = doc.createElement("Notes");
		e.appendChild( notes );
		txtNode = doc.createTextNode( entry.notes );
		notes.appendChild( txtNode );
	}

	docElem.appendChild( e );
  }

  QTextStream ts( &f );
  ts.setEncoding( QTextStream::UnicodeUTF8 );
  doc.save( ts, 0 );
}

DialerExt::DialerExt( QObject *targObject0, int fd0 ):
	Dialer( fd0 ), targObject(targObject0) {
}

void DialerExt::cancelDialing() {
  char c = 1;
  write( cmdPipe[1], &c, 1 );
}

void DialerExt::redial() {
  char c = 2;
  write( cmdPipe[1], &c, 1 );
}

void DialerExt::increaseTimeout() {
  char c = 3;
  write( cmdPipe[1], &c, 1 );
}

void DialerExt::decreaseTimeout() {
  char c =  4;
  write( cmdPipe[1], &c, 1 );
}

void DialerExt::dialing() {
  qApp->postEvent( targObject, new DialerEvent( DE_DIALING, dialingTimeout ));
}

void DialerExt::sleeping() {
  qApp->postEvent( targObject, new DialerEvent( DE_SLEEPING, dialingPause ) );
}

void DialerExt::cancellingDial() {
  qApp->postEvent( targObject, new DialerEvent( DE_WAIT, dialingPause ) );
}

DialerThread::DialerThread( const char *phone0, int fd0,
		QObject *eventReceiver0 ):
	eventReceiver(eventReceiver0), dialingDone(false), connected(false),
	de(0), fd(fd0) {
  QSettings s;
  s.setPath( "vr5", "term" );

  QString t = s.readEntry("term/dialingType");
  if ( !t.isEmpty() ) dialingType = t[0];

  phone = strdup( phone0 );
}

DialerThread::~DialerThread() {
  free( phone );
}

void DialerThread::run() {
  DialerExt dialer( eventReceiver, fd );
  if ( !dialingType.isEmpty() ) dialer.dialingType[0] = dialingType.latin1()[0];

  de = &dialer;
  connected = dialer.dial( phone );
  dialingDone = true;

  qApp->postEvent( eventReceiver, new QCustomEvent( 100 ) );
}

void DialerThread::cancelDialing() {
  de->cancelDialing();
}

void DialerThread::redial() {
  de->redial();
}

void DialerThread::increaseTimeout() {
  de->increaseTimeout();
}

void DialerThread::decreaseTimeout() {
  de->decreaseTimeout();
}

void DialDirectory::doDialing() {
  DirLVItem *it = (DirLVItem*)dirLV->selectedItem();
  if ( !it ) return;

  QWidget::hide();
  grabSerialLine();

  DialerDialog d( it->text(0), it->text(1), (QWidget*)parent() );
  connect( &d, SIGNAL(redialRequested()), SLOT(redial()) );
  connect( &d, SIGNAL(increaseTimeout()), SLOT(increaseTimeout()) );
  connect( &d, SIGNAL(decreaseTimeout()), SLOT(decreaseTimeout()) );

  dialer = new DialerThread( it->text(1), fd, &d );
  dialer->start();

  QEventLoop *loop = qApp->eventLoop();
  int mask = QEventLoop::AllEvents|QEventLoop::WaitForMore;

  while ( !dialer->dialingDone ) {
  	loop->processEvents( mask );
	if ( d.isHidden() && !dialer->dialingDone ) {
		dialer->cancelDialing();
		break;
	}
  }
  dialer->wait();
  delete dialer;

  releaseSerialLine();
  if ( dialer->connected ) {
  	dialList[it->listPos].calls++;
	dialList[it->listPos].lastLogin = QDate::currentDate();
	dialListModified = true;

	loop->exitLoop();	// hack
	reject();

	connectIdx = it->listPos;
  }
  else
  	QWidget::show();
}

void DialDirectory::redial() {
  dialer->redial();
}

void DialDirectory::increaseTimeout() {
  dialer->increaseTimeout();
}

void DialDirectory::decreaseTimeout() {
  dialer->decreaseTimeout();
}

void DialDirectory::editEntry() {
  int rc, idx;
  QString name, phone, notes, dlSuffix;
  bool iemsiLogin;
  QString iemsiLoginName, iemsiPassword;
  DirLVItem *it = (DirLVItem*)dirLV->selectedItem();
 
  if ( !it ) return;
  
  idx = it->listPos;

  EntryEditUI w( this );
  w.nameLE->setText( it->text(0) );
  w.phoneLE->setText( it->text(1) );
  w.useIemsiCB->setChecked( dialList[idx].iemsiLogin );
  w.iemsiLoginLE->setText( dialList[idx].iemsiLoginName );
  w.iemsiPasswdLE->setText( dialList[idx].iemsiPassword );
  w.dlSuffixLE->setText( dialList[idx].dlSuffix );
  w.notesTE->setText( dialList[idx].notes );
  
  rc = w.exec();
  if ( rc == QDialog::Rejected ) return;

  name = w.nameLE->text();
  phone = w.phoneLE->text();
  iemsiLogin = w.useIemsiCB->isChecked();
  iemsiLoginName = w.iemsiLoginLE->text();
  iemsiPassword = w.iemsiPasswdLE->text();
  dlSuffix = w.dlSuffixLE->text();
  notes = w.notesTE->text();
  it->setText( 0, name );
  it->setText( 1, phone );

  dialList[idx].name = name;
  dialList[idx].phone = phone;
  dialList[idx].iemsiLogin = iemsiLogin;
  dialList[idx].iemsiLoginName = iemsiLoginName;
  dialList[idx].iemsiPassword = iemsiPassword;
  dialList[idx].dlSuffix = dlSuffix;
  dialList[idx].notes = notes;
  dialListModified = true;
  
  updateNotesDisplay( it );
}

void DialDirectory::addEntry() {
  QListViewItem *it;
  QString name, phone, notes, dlSuffix;
  bool iemsiLogin;
  QString iemsiLoginName, iemsiPassword;
  int rc;

  EntryEditUI w( this );
  rc = w.exec();
  if ( rc == QDialog::Rejected ) return;

  name = w.nameLE->text();
  phone = w.phoneLE->text();
  iemsiLogin = w.useIemsiCB->isChecked();
  iemsiLoginName = w.iemsiLoginLE->text();
  iemsiPassword = w.iemsiPasswdLE->text();
  dlSuffix = w.dlSuffixLE->text();
  notes = w.notesTE->text();

  DialEntry entry;
  entry.name = name;
  entry.phone = phone;
  entry.iemsiLogin = iemsiLogin;
  entry.iemsiLoginName = iemsiLoginName;
  entry.iemsiPassword = iemsiPassword;
  entry.calls = 0;
  entry.dlSuffix = dlSuffix;
  entry.notes = notes;
  dialList.push_back( entry );

  it = new DirLVItem( dialList.size()-1, dirLV, dirLV->lastItem(),
  	name, phone, "0", "" );
  dirLV->setSelected( it, true );
  dirLV->setCurrentItem( it );
  updateNotesDisplay( it );
  dirLV->ensureItemVisible( it );

  dialListModified = true;
}

void DialDirectory::removeEntry() {
  DirLVItem *it = (DirLVItem*)dirLV->selectedItem();
  QListViewItem *next;
  int rc, pos;

  if ( !it ) return;

  rc = QMessageBox::warning( this, "Dialing Directory",
  	"Do you really want to remove record?",
	QMessageBox::Yes, QMessageBox::No );

  if ( rc == QMessageBox::No ) return;

  next = it->nextSibling();
  if ( !next ) {
  	next = dirLV->firstChild();
	while ( next && next->nextSibling() != it )
		next = next->nextSibling();
  }

  pos = it->listPos;
  delete it;
  dialList.erase( dialList.begin() + pos );

  if ( next ) {
	dirLV->setSelected( next, true );
	dirLV->setCurrentItem( next );
	dirLV->ensureItemVisible( next );
  }

  dialListModified = true;
}

void DialDirectory::updateNotesDisplay( QListViewItem *it0 ) {
  DirLVItem *it = (DirLVItem*)it0;
  if ( !it ) return;
  infoL->setText( dialList[it->listPos].notes );
}

void DialDirectory::keyPressEvent( QKeyEvent *e ) {
  if ( e->key() == Key_F4 ) {
	editEntry();
	return;
  }

  return DialDirectoryUI::keyPressEvent( e );
}

