#include <qdom.h>
#include <qfile.h>
#include <qlistview.h>
#include <qcombobox.h>
#include <qlineedit.h>
#include "term.h"
#include "design/portsetupui.h"
#include "moc_portprofile.h"

PortProfileList profiles;
int currentPortProfile;

static const char *flowConfText[] = { "none", "ctsrts", "xonxoff" };
static const char *flowItemText[] = { "None", "CTS/RTS", "XON/XOFF" };

void PortProfileList::load() {
  int i, currentFlag;
  PortProfile profile;
  QString fname;

  QDomDocument doc( "PortProfileList" );

  fname = ConfigManager::getFilename( "portprofile.xml" );
  QFile f( fname );

  if ( !f.open( IO_ReadOnly ) ) return;

  doc.setContent( &f );

  QDomNodeList nl;
  QDomElement docElem, e;

  docElem = doc.documentElement();
  nl = docElem.elementsByTagName( "Profile" );

  clear();

  currentPortProfile = -1;
  for ( i = 0; i < nl.count(); i++ ) {
  	e = nl.item(i).toElement();
	profile.name = e.attribute( "name" );
	profile.device = e.attribute( "device" );
	profile.speed = e.attribute( "speed" ).toInt();
	profile.bitset = e.attribute( "bitset" );
	
	QString flow = e.attribute( "flow", "none" );
	profile.flow = PortProfile::FLOW_NONE;
	for ( int j = 0; j < 2; j++ )
		if ( flow == flowConfText[j] 	) {
			profile.flow = j;
			break;
		}

	currentFlag = e.attribute( "current", "0" ).toInt();
	if ( currentFlag ) currentPortProfile = i;

	push_back( profile );
  }
}

void PortProfileList::save() {
  QDomDocument doc( "PortProfileList" );

  QString fname = ConfigManager::getFilename( "portprofile.xml" );
  QFile f( fname );

  if ( !f.open( IO_ReadWrite | IO_Truncate ) ) return;

  QDomElement docElem, e;
  docElem = doc.createElement( "PortProfileList" );
  doc.appendChild( docElem );

  for ( int i = 0; i < profiles.size(); i++ ) {
  	e = doc.createElement( "Profile" );
	e.setAttribute( "name", profiles[i].name );
	e.setAttribute( "device", profiles[i].device );
	e.setAttribute( "speed", profiles[i].speed );
	e.setAttribute( "bitset", profiles[i].bitset );
	e.setAttribute( "flow", flowConfText [profiles[i].flow] );
	if ( i == currentPortProfile )
		e.setAttribute( "current", "1" );
	docElem.appendChild( e );
  }

  QTextStream ts( &f );
  doc.save( ts, 0 );
}

int PortProfileList::speedCode( int itemPos ) {
  switch ( at(itemPos).speed ) {
  case 300: return B300;
  case 600: return B600;
  case 1200: return B1200;
  case 2400: return B2400;
  case 4800: return B4800;
  case 9600: return B9600;
  case 19200: return B19200;
  case 38400: return B38400;
  case 57600: return B57600;
  case 115200: return B115200;
  case 230400: return B230400;
  case 460800: return B460800;
  default: return B0;
  }
}

class ProfileSelectorItem: public QListViewItem {
public:
  int id;
  ProfileSelectorItem( int id0, QListView *lv, QListViewItem *after ):
	QListViewItem( lv, after ), id( id0 ) {}
};

PortProfileSelector::PortProfileSelector( QWidget *parent ):
	PortProfileUI( parent ) {
  int i;
  QListViewItem *it;

  setFixedSize( size() );

  it = 0;
  for ( i = 0; i < profiles.size(); i++ ) {
	it = new ProfileSelectorItem( i, lv, it );
	it->setText( 0, profiles[i].name );
	it->setText( 1, profiles[i].device );
  	it->setText( 2, QString::number( profiles[i].speed ) );
	it->setText( 3, profiles[i].bitset );
	it->setText( 4, flowItemText[profiles[i].flow] );

	if ( i == currentPortProfile ) {
		lv->setCurrentItem( it );
		lv->setSelected( it, true );
	}
  }

  show();
}

void PortProfileSelector::updateProfile( PortSetupUI *w,
	 PortProfile *profile, ProfileSelectorItem *it ) {

  profile->name = w->nameLE->text();
  profile->device = w->deviceLE->text();
  profile->speed = w->baudrateCB->currentText().toInt();
  profile->bitset = QString("%1,%2,%3").arg( w->bitsCB->currentText() )
	.arg( w->parityCB->currentText()[0].lower() )
	.arg( w->stopCB->currentText() ) ;
  profile->flow = w->flowCB->currentItem();

  it->setText( 0, profile->name );
  it->setText( 1, profile->device );
  it->setText( 2, QString::number( profile->speed ) );
  it->setText( 3, profile->bitset );
  it->setText( 4, flowItemText[profile->flow] );
}

void PortProfileSelector::editEntry() {
  int rc, id;
  char pc;
  QStringList opt;
  QString parity;

  ProfileSelectorItem *it = (ProfileSelectorItem*) lv->selectedItem();
  if ( !it ) return;

  id = it->id;

  opt = QStringList::split( ',', profiles[id].bitset );
  pc = opt[1].latin1()[0];
  if ( pc == 'n' ) parity = "None";
  else if ( pc == 'o' ) parity = "Odd";
  else if ( pc == 'e' ) parity = "Even";

  PortSetupUI w( this );
  w.nameLE->setText( profiles[id].name );
  w.deviceLE->setText( profiles[id].device );
  w.baudrateCB->setCurrentText( QString::number( profiles[id].speed ) );
  w.bitsCB->setCurrentText( opt[0] );
  w.stopCB->setCurrentText( opt[2] );
  w.parityCB->setCurrentText( parity );
  w.flowCB->setCurrentText( flowItemText[profiles[id].flow] );

  rc = w.exec();
  if ( rc == QDialog::Rejected ) return;

  updateProfile( &w, &profiles[id], it );
}

void PortProfileSelector::addEntry() {
  int rc, pos;
  ProfileSelectorItem *it;

  PortSetupUI w( this );

  rc = w.exec();
  if ( rc == QDialog::Rejected ) return;

  pos = profiles.size();
  profiles.push_back( PortProfile() );

  it = new ProfileSelectorItem( pos, lv, lv->lastItem() );

  updateProfile( &w, &profiles[pos], it );

  lv->setSelected( it, true );
  lv->setCurrentItem( it );
  lv->ensureItemVisible( it );
}

void PortProfileSelector::removeEntry() {
  QListViewItem *next;

  ProfileSelectorItem *it = (ProfileSelectorItem*) lv->selectedItem();
  if ( !it ) return;

  int id = it->id;
  next = it->nextSibling();
  if ( !next ) {
	next = lv->firstChild();
	while ( next && next->nextSibling() != it )
		next = next->nextSibling();
  }

  delete it;
  profiles.erase( profiles.begin() + id );

  if ( next ) {
	lv->setSelected( next, true );
	lv->setCurrentItem( next );
	lv->ensureItemVisible( next );
  }
}

int PortProfileSelector::getSelectedProfileId() {
  ProfileSelectorItem *it = (ProfileSelectorItem*) lv->selectedItem();
  if ( !it ) return -1;

  currentPortProfile = it->id;

  return currentPortProfile;
}
