#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <unistd.h>
#include <sys/ioctl.h>

#include <functional>

#include <qapplication.h>
#include <qclipboard.h>
#include <qsettings.h>
#include <qtimer.h>
#include <qpainter.h>
#include <qsocketnotifier.h>
#include <qtextcodec.h>
#include <qpixmap.h>
#include <qthread.h>
#include <qeventloop.h>
#include <qdialog.h>
#include <qlistview.h>
#include <qlabel.h>
#include <qfile.h>
#include <qfiledialog.h>
#include <qradiobutton.h>
#include <qpushbutton.h>

#include "dialdirectory.h"
#include "crc/crctab.h"
extern "C" {
#include "xmodem/xmodem.h"
}
#include "zmodem/zmodem.h"
#include "transferdialog.h"
#include "zmodemext.h"

#include "moc_term.h"
#include "moc_serialline.h"
#include "portprofile.h"
#include "statusline.h"
#include "design/selectprotodlg.h"

#include "term-dbus/d-xmodem.h"

SerialLine::SerialLine(): serial_fd(-1), lineName(0), reopenTimer(0) {
}

SerialLine::~SerialLine() {
  if ( serial_fd != -1 ) {
//  	tcflow( serial_fd, TCOOFF );
//	tcflow( serial_fd, TCIOFF );
	tcflush( serial_fd, TCIOFLUSH );
	close( serial_fd );
  }
  free( lineName );
}

int SerialLine::open( const char *lineName0, int speed0, int flow0 ) {
  free( lineName );
  lineName = strdup( lineName0 );
  speed = speed0;
  flow = flow0;

  serial_fd = ::open( lineName, O_RDWR|O_NONBLOCK|O_NOCTTY );
  if ( serial_fd == -1 ) return -1;
  
  termios ts;
  tcgetattr( serial_fd, &ts );
  cfmakeraw( &ts );
  if ( flow == PortProfile::FLOW_CTSRTS ) {
	ts.c_iflag &= ~IXON;
	ts.c_cflag |= CRTSCTS;
  }
  else if ( flow == PortProfile::FLOW_XONXOFF ) {
	ts.c_iflag |= IXON;
	ts.c_cflag &= ~CRTSCTS;
  }
  else {
	ts.c_iflag &= ~IXON;
	ts.c_cflag &= ~CRTSCTS;
  }
  cfsetospeed( &ts, speed );
  cfsetispeed( &ts, B0 );
  tcsetattr( serial_fd, TCSANOW, &ts );

  return 0;
}

void SerialLine::reopen( const char *lineName0, int speed0, int flow0 ) {
  if ( serial_fd != -1 ) {
	tcflow( serial_fd, TCOOFF );
	tcflow( serial_fd, TCIOFF );
	tcflush( serial_fd, TCIOFLUSH );
	close( serial_fd );
  }

  serial_fd = open( lineName0, speed0, flow0 );
  if ( serial_fd != -1 ) {
	if ( reopenTimer ) {
		reopenTimer->stop();
		reopenTimer->deleteLater();
		reopenTimer = 0;
	}
	emit lineReopened( this );
  }
  else if ( !reopenTimer ) {
	reopenTimer = new QTimer( this );
	connect( reopenTimer, SIGNAL(timeout()), SLOT(reopen()) );
	reopenTimer->start( 1000 );
  }
}

void SerialLine::reopen() {
  char *tmp = strdup( lineName );
  reopen( tmp, speed, flow );
  free( tmp );
}

int SerialLine::fd() {
  return serial_fd;
}

ZModemExt::ZModemExt( QWidget *targWidget, int fd ):
  ZModem( fd, fd ), target(targWidget) {
}

void ZModemExt::skipFile() {
  char c = 1;
  write( cmdPipe[1], &c, 1 );
}

void ZModemExt::cancelTransfer() {
  char c = 2;
  write( cmdPipe[1], &c, 1 );
}

void ZModemExt::feFileInfo( char *fname, int bytes ) {
  XferEvent *zme = new XferEvent( XFER_FILE_INFO );
  zme->fname = strdup( fname );
  zme->bytes = bytes;
  qApp->postEvent( target, zme );
}

void ZModemExt::feFileResume( int startPos ) {
  XferEvent *zme = new XferEvent( XFER_FILE_RESUME );
  zme->bytes = startPos;
  qApp->postEvent( target, zme );
}

void ZModemExt::feFileData( int bytes ) {
  XferEvent *zme = new XferEvent( XFER_FILE_DATA );
  zme->bytes = bytes;
  qApp->postEvent( target, zme );
}

void ZModemExt::feFileDone() {
  XferEvent *zme = new XferEvent( XFER_FILE_DONE );
  qApp->postEvent( target, zme );
}

bool ZModemExt::stopRxRequested() {
  return false;
}

class ScrollbackBuffer {
public:
  int xsize, ysize;
  Cell *data, *first;
  int lines;

  ScrollbackBuffer( int xsize0, int ysize0 ): xsize(xsize0), ysize(ysize0), lines(0) {
	first = data = new Cell[xsize * ysize];
  }
  ~ScrollbackBuffer() {
  	delete[] data;
  }

  void append( Cell *line ) {
	if ( lines < ysize ) {
		memcpy( first + lines * xsize, line, xsize * sizeof(Cell) );
		lines++;
	}
	else {
  		memcpy( first, line, xsize * sizeof(Cell) );

		first += xsize;
		if ( first >= data + xsize * ysize ) first = data;
	}
  }

  Cell *line( int n ) {
	if ( lines < ysize )
		return first + (lines - n - 1) * xsize;
	else {
		Cell *p = first - (n + 1) * xsize;
		return p >= data? p: p + xsize * ysize;
	}
  }
};

TermEmu::TermEmu( QWidget *parent, const QFont& fnt0, int xsize0, int ysize0 ):
	QWidget( parent ), fnt(fnt0), xsize(xsize0), ysize(ysize0),
	sl(0), cursor(0), cursor_on(false),
	selectionStart(-1), selectionEnd(-1), hasSelection(false), selecting(false),
	scrollbackLines(0), ansi_state( ANSI_TEXT ),
	bright_color(false), reverse_color(false), saved_pos(0), lineNotifier(0),
	connInfoValid(false),
	iemsiState( IEMSI_WAIT_IRQ ), iemsiData(0), iemsiDataLen(0) {

  sbb = new ScrollbackBuffer( xsize, ysize * 2 );
  codec = QTextCodec::codecForName( "IBM 866" );

  col_pal[0] = black;
  col_pal[1] = red.dark();
  col_pal[2] = green.dark();
  col_pal[3] = yellow.dark();
  col_pal[4] = blue.dark();
  col_pal[5] = magenta.dark();
  col_pal[6] = cyan.dark();
  col_pal[7] = gray;
  col_pal[8] = gray.dark();
  col_pal[9] = red;
  col_pal[10] = green;
  col_pal[11] = yellow;
  col_pal[12] = blue;
  col_pal[13] = magenta;
  col_pal[14] = cyan;
  col_pal[15] = white;

  bgcolor = 7;
  fgcolor = 0;
  curs_color = 0;

  QFontMetrics fm( fnt );
  cellw = fm.width( "A" );
  cellh = fm.height();
//  printf( "height=%d, ascent=%d, descent=%d\n",
//  	fm.height(), fm.ascent(), fm.descent() );

  resize( xsize * cellw, ysize * cellh );

  setFocusPolicy( StrongFocus );

  data = new Cell[xsize * ysize];
  for ( int i = 0; i < xsize * ysize; i++ )
  	outp_at( i, ' ' );

  
  QSettings s;
  s.setPath( "vr5", "term" );

  if ( s.readEntry( "term/linelogEnable", "" ) == "yes" )
	logfile = open( ConfigManager::getVarPath("line.log"), O_WRONLY|O_CREAT|O_APPEND, 0644 );
  else
	logfile = -1;

  initstr = s.readEntry( "term/initstr" ).utf8();

  QTimer *t = new QTimer( this );
  connect( t, SIGNAL(timeout()), SLOT(drawCursor()) );
  t->start(500);

  memset( zmBuf, 0, sizeof zmBuf );
  memset( iemsiBuf, 0, sizeof iemsiBuf );
  dlPreffix = ConfigManager::getVarPath("files");
}

TermEmu::~TermEmu() {
  delete sbb;
  delete[] data;
  if ( logfile != -1 ) ::close( logfile );
}

void TermEmu::outp_at( int pos, const QChar& c ) {
  data[pos].c = c;
  if ( reverse_color ) {
	data[pos].fg = bgcolor;
	data[pos].bg = fgcolor;
  }
  else {
	data[pos].fg = fgcolor;
	data[pos].bg = bgcolor;
  }
}

void TermEmu::attachLine( SerialLine *line ) {
  sl = line;
  connect( sl, SIGNAL(lineReopened(SerialLine*)), SLOT(lineReopened(SerialLine*)) );

  if ( sl->fd() == -1 ) {
	statLine->disable();
  	sl->reopen();
  }
  else {
  	delete lineNotifier;

	lineNotifier = new QSocketNotifier( sl->fd(), QSocketNotifier::Read, this );
	connect( lineNotifier, SIGNAL(activated(int)), SLOT(fetchFromLine(int)) );
  }
}

void TermEmu::disconnectLine() {
  delete lineNotifier;
  lineNotifier = 0;
}

void TermEmu::fetchFromLine( int fd ) {
  unsigned char buf[100];
  int rc;

  rc = read( fd, buf, sizeof buf );
  if ( rc <= 0 ) {
  	puts( "hangup, reopening line" );

	if ( connInfoValid ) {
		log( "disconnected" );
		connInfoValid = false;
	}

	delete lineNotifier;
	lineNotifier = 0;

	statLine->disable();
	sl->reopen();
  }
  else if ( rc > 0 ) {
  	write( logfile, buf, rc );
  	addStr( buf, rc );
  }
}

void TermEmu::lineReopened( SerialLine *sl ) {
  delete lineNotifier;
  lineNotifier = new QSocketNotifier( sl->fd(), QSocketNotifier::Read, this );
  connect( lineNotifier, SIGNAL(activated(int)), SLOT(fetchFromLine(int)) );
}

QChar TermEmu::translateChar( unsigned char c ) {
  unsigned short ctl_table[32] = {
  	  0x20,    0x1,  0x2, 0x2665, 0x2666, 0x2663, 0x2660,    0x7,
	   0x8,    0x9,  0xa,    0xb,    0xc,    0xd,    0xe,    0xf,
	0x25b6, 0x25c0, 0x12,   0x13,   0x14,   0x15,   0x16,   0x17,
	0x2191,   0x19, 0x1a,   0x1b,   0x1c, 0x2194, 0x25b2, 0x25bc,
  };
  QString s;
  
  if ( c < 0x20 ) {
  	if ( ctl_table[c] == c )
		printf( "displaying char %d\n", c );
  	return QChar( ctl_table[c] );
  }
  else {
	if ( codec ) s = codec->toUnicode( (char*)&c, 1 );
	else s = QString::fromLocal8Bit( (char*)&c, 1 );
  }

  return s[0];
}

// interpret \e[1;3xm as foreground color setup, where x is:
// 0 for black,
// 1 for red,
// 2 for green,
// 3 for yellow,
// 4 for blue,
// 5 for magenta,
// 6 for cyan,
// 7 for white

void TermEmu::setFgColor( int color_idx ) {
  fgcolor = color_idx;
  if ( bright_color ) fgcolor += 8;
}

void TermEmu::process_m_cmd() {
  int i;
 
  for ( i = 0; i <= num_idx; i++ )
	switch( num[i] ) {
	case 0:
		fgcolor = 0;
		bgcolor = 7;
		bright_color = false;
		reverse_color = false;
		break;
	case 1: 
		if ( fgcolor < 8 ) fgcolor += 8;
		bright_color = true;
		break;
	
	case 22:
		if ( fgcolor >= 8 ) fgcolor -= 8;
		bright_color = false;
		break;
	
	case 7:
		reverse_color = true;
		break;
	
	case 27:
		reverse_color = false;
		break;

	case 30: setFgColor( 0 ); break;
	case 31: setFgColor( 1 ); break;
	case 32: setFgColor( 2 ); break;
	case 33: setFgColor( 3 ); break;
	case 34: setFgColor( 4 ); break;
	case 35: setFgColor( 5 ); break;
	case 36: setFgColor( 6 ); break;
	case 37: setFgColor( 7 ); break;
	case 39: setFgColor( 0 );break;

	case 40: bgcolor = 0; break;
	case 41: bgcolor = 1; break;
	case 42: bgcolor = 2; break;
	case 43: bgcolor = 3; break;
	case 44: bgcolor = 4; break;
	case 45: bgcolor = 5; break;
	case 46: bgcolor = 6; break;
	case 47: bgcolor = 7; break;
	case 49: bgcolor = 7; break;
	}
}

void TermEmu::process_n_cmd() {
  char buf[20];

  if ( num[0] != 6 ) return;

  sprintf( buf, "\033[%d;%dR", cursor / xsize + 1, cursor % xsize + 1 );
  write( sl->fd(), buf, strlen(buf) );
}

void TermEmu::moveCursor( int new_pos ) {
  cursor_on = false;
  repaintCell( cursor );
  cursor = new_pos;
  cursor_on = true;
  repaintCell( cursor );
}

enum {
  MOVE_UP,
  MOVE_DOWN,
  MOVE_RIGHT,
  MOVE_LEFT
};

void TermEmu::repMoveCursor( int dir ) {
  int i, mcnt = num[0];
  int pos = cursor;
 
  if ( mcnt < 1 ) mcnt = 1;

  for ( i = 0; i < mcnt; i++  )
	switch ( dir ) {
	case MOVE_UP:
		if ( pos < xsize ) goto move;
		pos -= xsize;
		break;
	case MOVE_DOWN:
		pos += xsize;
		moveCursor( pos );
		fixCursor();
		break;
	case MOVE_RIGHT:
		if ( pos % xsize == xsize - 1 ) goto move;
		pos++;
		break;
	case MOVE_LEFT:
		if ( pos % xsize == 0 ) goto move;
		pos--;
		break;
	}
move:
  if ( cursor != pos ) moveCursor( pos );
}

void TermEmu::process_A_cmd() {
  repMoveCursor( MOVE_UP );
}

void TermEmu::process_B_cmd() {
//  printf( "B MOVE_DOWN %d (%d;%d)\n", num[0], cursor%xsize, cursor/xsize );
  repMoveCursor( MOVE_DOWN );
//  printf( "B done (%d;%d)\n", cursor%xsize, cursor/xsize );
}

void TermEmu::process_C_cmd() {
  repMoveCursor( MOVE_RIGHT );
}

void TermEmu::process_D_cmd() {
  repMoveCursor( MOVE_LEFT );
}

void TermEmu::process_G_cmd() {
  int newx = num[0]-1;
  if ( newx < 0 ) newx = 0;

  moveCursor( cursor - cursor % xsize + newx );
}

void TermEmu::eraseScreenPart( int start, int end ) {
  for ( int i = start; i < end; i++ ) outp_at( i, ' ' );

  if ( hasSelection &&
	selectionStart >= start || selectionEnd > start &&
	selectionStart <= end || selectionEnd < end )
		hasSelection  = false;

  repaint( false );
}

void TermEmu::process_J_cmd() {
  switch ( num[0] ) {
  case 0:
	eraseScreenPart( cursor, xsize * ysize );
	break;
  case 1:
	eraseScreenPart( 0, cursor );
	break;
  case 2:
	cursor = 0;
	eraseScreenPart( 0, xsize * ysize );
	break;
  }
}

void TermEmu::eraseLinePart( int pos, int count ) {
  int i;

  cursor_on = false;
  repaintCell( cursor );

  for ( i = 0; i < count; i++ ) {
  	outp_at( pos, ' ' );
	cellChanged( pos );
	pos++;
  }
  
  cursor_on = true;
  repaintCell( cursor );
}

void TermEmu::process_K_cmd() {
  int i, pos, count;
  switch ( num[0] ) {
  case 0:
  	eraseLinePart( cursor, xsize - cursor % xsize );
  	break;
  case 1:
  	eraseLinePart( cursor - cursor % xsize, cursor % xsize );
  	break;
  case 2:
  	eraseLinePart( cursor - cursor % xsize, xsize );
	break;
  }
}

void TermEmu::process_H_cmd() {
  int newx, newy;

  if ( num_idx >= 1 ) {
  	newy = num[0]-1;
	newx = num[1]-1;
	if ( newy < 0 ) newy = 0;
	if ( newx < 0 ) newx = 0;
  }
  else
  	newx = newy = 0;

//  printf( "H%d;%d move cursor (%d;%d)\n",
//  	newx, newy, cursor%xsize, cursor/xsize );
  moveCursor( newy*xsize + newx );
  fixCursor();
//  printf( "H done (%d;%d)\n", cursor%xsize, cursor/xsize );
}

void TermEmu::process_P_cmd() {
  int char_count = num[0];
  if ( char_count > xsize - cursor % xsize ) char_count = xsize - cursor % xsize;

  int move_count = xsize - cursor % xsize - char_count;
  memmove( &data[cursor], &data[cursor + char_count], sizeof(*data) * move_count );
  for ( int i = 0; i < move_count; i++ ) cellChanged( cursor  + i );

  eraseLinePart( cursor + move_count, xsize - (cursor + move_count) % xsize );
}

void TermEmu::process_X_cmd() {
  eraseLinePart( cursor, num[0] );
}

void TermEmu::process_d_cmd() {
  int newy = num[0]-1;
  if ( newy < 0 ) newy = 0;

  moveCursor( newy * xsize + cursor % xsize );
  fixCursor();
}

class XFerThread: public QThread {
public:
  bool sessionTerminated;

  XFerThread( bool rx0 ): rx(rx0), sessionTerminated(false)
	 {}
  void setDownloadPath( const QString& dlPath0 ) {
  	dlPath = dlPath0;
  }
  void setTxList( const QStringList& txList0 ) {
	txList = txList0;
  }
  virtual void setEventReceiver( QWidget *receiver ) = 0;
  virtual void skipFile() = 0;
  virtual void cancelTransfer() = 0;
protected:
  bool rx;
  QString dlPath;
  QStringList txList;
};

class XModemThread: public XFerThread {
public:
  static int fd;
  static QWidget *xferDlg;
  QString rxName;
  XferProto xferProto;

  XModemThread( bool rx0, int fd0 ): XFerThread( rx0 ) {
	fd = fd0;
  }

  void setEventReceiver( QWidget *receiver ) {
    xferDlg = receiver;
  }

  void skipFile() {
	XModem_cancelled = 1;
  }

  void cancelTransfer() {
	XModem_cancelled = 1;
  }

  void run() {
	XModem_feFileInfo = feFileInfo;
	XModem_feFileData = feFileData;
	XModem_feFileDone = feFileDone;
	XModem_cancelled  = 0;

	if ( rx ) {
		QByteArray fname = rxName.local8Bit();
		xmodemReceive( fname );
	}
	else {
		QByteArray fname = txList[0].local8Bit();
		if ( xferProto == XferXmodem )
			xmodemTransmit128( fname );
		else if ( xferProto == XferXmodem1k )
			xmodemTransmit1k( fname );
	}

	sessionTerminated = true;
	qApp->postEvent( xferDlg, new QCustomEvent( XFER_SESSION_TERM ) );
  }

  static void feFileInfo( const char *fname0, int bytes ) {
	XferEvent *zme = new XferEvent( XFER_FILE_INFO );

	const char *fname = strrchr( fname0, '/' );
	if ( fname ) fname++; else fname = fname0;

	zme->fname = strdup( fname );
	zme->bytes = bytes;
	qApp->postEvent( xferDlg, zme );
  }

  static void feFileData( int bytes ) {
	XferEvent *zme = new XferEvent( XFER_FILE_DATA );
	zme->bytes = bytes;
	qApp->postEvent( xferDlg, zme );
  }

  static void feFileDone() {
	XferEvent *zme = new XferEvent( XFER_FILE_DONE );
	qApp->postEvent( xferDlg, zme );
  }
};

int XModemThread::fd = -1;
QWidget *XModemThread::xferDlg = 0;

extern "C"
int XModem_inbyte( int timeout ) {
  unsigned char c;
  int rc;

  for (;;) {
        rc = read( XModemThread::fd, &c, 1 );
        if ( rc ==  1 ) return  c;
        if ( rc == -1 && errno != EAGAIN ) return -1;

        usleep( 1000 );
        if ( timeout && !--timeout ) return -2;
  }
}

extern "C"
int XModem_inbytes( unsigned char *dest, int count, int timeout0 ) {
  unsigned char c;
  int rc;
  int timeout = timeout0;

  for (;;) {
        rc = read( XModemThread::fd, dest, count );
        if ( rc > 0 ) {
		dest  += rc;
		count -= rc;
		if ( !count ) return 1;

		timeout = timeout0;
		continue;
	}

	if ( rc == -1 && errno != EAGAIN ) return -1;
		
       	usleep( 1000 );
        if ( timeout && !--timeout ) return -2;
  }
}

extern "C"
void XModem_outbyte( int c ) {
  while ( write( XModemThread::fd, &c, 1 ) == 0 ) usleep( 1000 );
}

extern "C"
void XModem_outbytes( unsigned char *src, int count ) {
  int rc;
  for (;;) {
  	rc = write( XModemThread::fd, src, count );
	if ( rc > 0 ) {
		src   += rc;
		count -= rc;
		if ( !count ) return;
	}

	if ( rc == -1 && errno != EAGAIN ) return;

	usleep( 1000 );
  }
}

extern "C"
void XModem_discardoutput() {
  tcflush( XModemThread::fd, TCOFLUSH );
}

class ZModemThread: public XFerThread {
public:
  ZModemExt *zm;

  ZModemThread( bool rx0, int fd0 ): XFerThread(rx0), zm(0), fd(fd0) {}

  void setEventReceiver( QWidget *receiver ) {
    eventReceiver = receiver;
  }

  void skipFile() {
	zm->skipFile();
  }

  void cancelTransfer() {
	zm->cancelTransfer();
  }

protected:
  QWidget *eventReceiver;
  int fd;
  virtual void run();
};

void ZModemThread::run() {
  zm = new ZModemExt( eventReceiver, fd );

  if ( rx ) {
	if ( !dlPath.isNull() ) {
  		free( zm->dlpath );
		zm->dlpath = strdup( dlPath.local8Bit().data() );
	}

	zm->receiveFiles( true );
  }
  else {
	char **flist;
	int count, i;

	count = txList.count();
	flist = new char*[count + 1];

	for ( i = 0; i < count; i++ )
		flist[i] = strdup( txList[i].local8Bit() );
	flist[i] = 0;

	zm->sendFiles( flist );

	for ( i = 0; i < count; i++ )
		free( flist[i] );
	delete[] flist;
  }

  sessionTerminated = true;
  qApp->postEvent( eventReceiver, new QCustomEvent( XFER_SESSION_TERM ) );

  delete zm;
}

void TermEmu::doXfer( XFerThread *xt, const QString& progressCaption ) {
  TransferDialog d( this );
  if ( !progressCaption.isEmpty() ) d.setCaption( progressCaption );

  connect( &d, SIGNAL(fileTransferStart(XferEvent*)),
	SLOT(processFileRxStart(XferEvent*)) );

  connect( &d, SIGNAL(fileTransferResume(XferEvent*)),
	SLOT(processFileRxResume(XferEvent*)) );

  connect( &d, SIGNAL(fileTransferDone()), SLOT(processFileRxDone()) );

  connect( &d, SIGNAL(skipFileRequested()),      SLOT(processXferSkipRequest()) );
  connect( &d, SIGNAL(cancelSessionRequested()), SLOT(processXferStopRequest()) );

  xt->setEventReceiver( &d );
  xt->start();

  QEventLoop *loop = qApp->eventLoop();
  int eventMask = QEventLoop::AllEvents|QEventLoop::WaitForMore;

  bool cancellingTransfer = false;
  transferSkipRequested = false;
  transferStopRequested = false;
  while ( !xt->sessionTerminated ) {
	loop->processEvents( eventMask );
	if ( !cancellingTransfer && transferStopRequested && !xt->sessionTerminated ) {
		xt->cancelTransfer();
		cancellingTransfer = true;
	}
	if ( transferSkipRequested ) {
		xt->skipFile();
		transferSkipRequested = false;
	}
  }

  xt->wait();
}

void TermEmu::processXferSkipRequest() {
  transferSkipRequested = true;
  qApp->postEvent( this, new QEvent( QEvent::User ) ); // for breaking out of event processing loop
}

void TermEmu::processXferStopRequest() {
  transferStopRequested = true;
  qApp->postEvent( this, new QEvent( QEvent::User ) );
}

void TermEmu::rxFileXModem( const QString& fname ) {
  releaseLine();

  XModemThread xt( true, sl->fd() );
  xt.rxName = fname;

  doXfer( &xt );

  grabLine();
}

void TermEmu::rxFileXModem() {
  QString dlPath = dlPreffix;
  if ( connInfoValid && !connInfo.dlSuffix.isNull() )
	dlPath += "/" + connInfo.dlSuffix;

  QString fname = QFileDialog::getSaveFileName( dlPath, QString::null, this );
  if ( !fname.isEmpty() ) rxFileXModem( fname );
}

void TermEmu::detectZModem( unsigned char c ) {
  const char *zmPattern = "**\030B00";
  pid_t child_pid;

  zmBuf[5] = c;
  if ( memcmp( zmBuf, zmPattern, 6 ) ) {
  	memmove( zmBuf, zmBuf+1, 5 );
	return;
  }

  releaseLine();
  fcntl( sl->fd(), F_SETFL, O_RDWR|O_NOCTTY );

  ZModemThread zt( true, sl->fd() );
  QString dlPath = dlPreffix;
  if ( connInfoValid && !connInfo.dlSuffix.isNull() )
	dlPath += "/" + connInfo.dlSuffix;
  zt.setDownloadPath( dlPath );

  doXfer( &zt, "" );

  fcntl( sl->fd(), F_SETFL, O_RDWR|O_NOCTTY|O_NONBLOCK );
  grabLine();
	
  memset( zmBuf, 0, sizeof zmBuf );
}

void TermEmu::txFilesXModem( const QStringList& files ) {
  releaseLine();

  XModemThread xt( false, sl->fd() );
  xt.xferProto = xferProto;
  xt.setTxList( files );

  doXfer( &xt, "Uploading files" );

  grabLine();
}

void TermEmu::txFilesXModem1k( const QStringList& files ) { txFilesXModem( files ); }
void TermEmu::txFilesYModem  ( const QStringList& files ) { txFilesXModem( files ); }

void TermEmu::txFileXModem( const QString& files ) {
  xferProto = XferXmodem;
  txFilesXModem( files );
}

class XModemQueueEvent: public QCustomEvent {
public:
  enum {
  	RxFile = QEvent::User + 1,  // User+0 used for doXfer cancellation
	TxFile
  };
  QString fname;

  XModemQueueEvent( int type, const QString& fname0 ): QCustomEvent(type), fname(fname0) {}
};

void TermEmu::queueRxXModem( const QString& fname ) {
  qApp->postEvent( this, new XModemQueueEvent( XModemQueueEvent::RxFile, fname ) );
}

void TermEmu::queueTxXModem( const QString& fname ) {
  qApp->postEvent( this, new XModemQueueEvent( XModemQueueEvent::TxFile, fname ) );
}

void TermEmu::customEvent( QCustomEvent *e ) {
  XModemQueueEvent *ev = (XModemQueueEvent*)e;

  if ( ev->type() == (int)XModemQueueEvent::RxFile ) rxFileXModem( ev->fname );
  else if ( ev->type() == (int)XModemQueueEvent::TxFile) txFilesXModem( ev->fname );
}

void TermEmu::txFilesZModem( const QStringList& files ) {
  releaseLine();
  fcntl( sl->fd(), F_SETFL, O_RDWR|O_NOCTTY );

  ZModemThread zt( false, sl->fd() );
  zt.setTxList( files );

  doXfer( &zt, "Uploading files" );

  fcntl( sl->fd(), F_SETFL, O_RDWR|O_NOCTTY|O_NONBLOCK );
  grabLine();
}

void TermEmu::detectIEmsi( unsigned char c ) {
  const char *iemsiPattern = "**EMSI_";
  const char *s;

  iemsiBuf[13] = c;
  if ( !memcmp(iemsiBuf, iemsiPattern, 7) )
  	puts( (char*)iemsiBuf );

  if ( !connInfoValid || !connInfo.iemsiLogin )
	goto skip_iemsi;

  switch ( iemsiState ) {
  case IEMSI_WAIT_IRQ:
	if ( !memcmp( iemsiBuf, "**EMSI_IRQ8E08", 14 ) ) {
send_ici:
		QString qdata = "{" + connInfo.iemsiLoginName +
			"}{}{}{}{}{" + connInfo.iemsiPassword +
			"}{}{}{}{}{}{Term,1.1}{}";
		qdata = "EMSI_ICI" + QString().sprintf( "%04X", qdata.length() ) + qdata;
		const char *s = qdata.latin1();
		int crc = 0xffffffff, len = strlen(s);

		for ( int i = 0; i < len; i++ )
			crc = updcrc32( *s++, crc );

		qdata = "**" + qdata + + QString().sprintf( "%08X", crc ) + "\r";
		s = qdata.latin1();
		write( sl->fd(), s, strlen(s) );
		iemsiState = IEMSI_WAIT_ISI;
	}
	break;
  case IEMSI_WAIT_ISI:
	if ( !memcmp( iemsiBuf, "**EMSI_ISI", 10 ) ) {
		int iemsiIsiLen = strtoul( (char*)iemsiBuf + 10, 0, 16 ) + 20;
		if ( iemsiDataLen < iemsiIsiLen ) {
			iemsiData = (char*)realloc( iemsiData, iemsiIsiLen );
			iemsiDataLen = iemsiIsiLen;
		}
		memcpy( iemsiData, iemsiBuf + 2, 12 );
		iemsiDataIdx = 12;

		iemsiState = IEMSI_COLLECT_ISI;
	}
	else if ( !memcmp( iemsiBuf, "**EMSI_NAKEEC3", 14 ) )
		goto send_ici;
	break;
  case IEMSI_COLLECT_ISI:
	iemsiData[iemsiDataIdx++] = c;

	if ( iemsiDataIdx == iemsiDataLen ) {
		s = iemsiData;
		int crc = 0xffffffff;

		for ( int i = 0; i < iemsiDataLen - 8; i++ )
			crc = updcrc32( *s++, crc );

		if ( strtoul( iemsiData + iemsiDataLen - 8, 0, 16 ) == crc ) {
			s = "**EMSI_ACKA490\r";
			write( sl->fd(), s, strlen(s) );
			write( sl->fd(), s, strlen(s) );
			iemsiState = IEMSI_WAIT_IRQ;
		}
		else {
			s = "**EMSI_NAKEEC3\r";
			write( sl->fd(), s, strlen(s) );
			iemsiState = IEMSI_WAIT_ISI;
		}
	}
	break;
  }

skip_iemsi:
  if ( !memcmp(iemsiBuf, iemsiPattern, 7) )
  	memset( iemsiBuf, 0, sizeof iemsiBuf );
  else
	memmove( iemsiBuf, iemsiBuf+1, sizeof(iemsiBuf) - 1 );
}

void TermEmu::grabLine() {
  if ( lineNotifier ) lineNotifier->setEnabled( true );
}

void TermEmu::releaseLine() {
  if ( lineNotifier ) lineNotifier->setEnabled( false );
}

void TermEmu::addStr( const unsigned char *str, int len ) {
  int i;
  unsigned char c;

  for ( i = 0; i < len; i++ ) {
  	c = str[i];
	detectZModem( c );
	detectIEmsi( c );

	switch ( ansi_state ) {
	case ANSI_ESC:
		if ( c == '[' ) {
			num_idx = 0;
			num[0] = 0;
			ansi_state = ANSI_GET_NUM1;
		}
		else ansi_state = ANSI_TEXT;
		break;
	case ANSI_GET_NUM1:
	  	if ( c == 'm' ) {
			process_m_cmd();
			ansi_state = ANSI_TEXT;
		}
	  	else if ( c == 'n' ) {
			process_n_cmd();
			ansi_state = ANSI_TEXT;
		}
		else if ( c == 's' ) {
//			printf( "S saving position (%d;%d)\n",
//				cursor%xsize, cursor/xsize );
			saved_pos = cursor;
			ansi_state = ANSI_TEXT;
		}
		else if ( c == 'u' ) {
//			printf( "U restoring at position (%d;%d)\n",
//				saved_pos%xsize, saved_pos/xsize );
			moveCursor( saved_pos );
			ansi_state = ANSI_TEXT;
		}
		else if ( c == 'A' ) {
			process_A_cmd();
			ansi_state = ANSI_TEXT;
		}
		else if ( c == 'B' ) {
			process_B_cmd();
			ansi_state = ANSI_TEXT;
		}
		else if ( c == 'C' ) {
			process_C_cmd();
			ansi_state = ANSI_TEXT;
		}
		else if ( c == 'D' ) {
			process_D_cmd();
			ansi_state = ANSI_TEXT;
		}
		else if ( c == 'G' ) {
			process_G_cmd();
			ansi_state = ANSI_TEXT;
		}
		else if ( c == 'J' ) {
			process_J_cmd();
			ansi_state = ANSI_TEXT;
		}
		else if ( c == 'K' ) {
			process_K_cmd();
			ansi_state = ANSI_TEXT;
		}
		else if ( c == 'H' || c == 'f' ) {
			process_H_cmd();
			ansi_state = ANSI_TEXT;
		}
		else if ( c == 'P' ) {
			process_P_cmd();
			ansi_state = ANSI_TEXT;
		}
		else if ( c == 'X' ) {
			process_X_cmd();
			ansi_state = ANSI_TEXT;
		}
		else if ( c == 'd' ) {
			process_d_cmd();
			ansi_state = ANSI_TEXT;
		}
	  	else if ( c == ';' ) {
			num_idx++;
			if ( num_idx < 10 )
				num[num_idx] = 0;
			else
				ansi_state = ANSI_TEXT;
		}
		else if ( isdigit(c) )
			num[num_idx] = num[num_idx] * 10 + c - '0';
		else {
			printf( "unknown cmd %c\n", c );
			ansi_state = ANSI_TEXT;
		}
		break;
	case ANSI_TEXT:
		if ( c == 033 )
			ansi_state = ANSI_ESC;
		else {
	  		if ( c == '\r' ) {
				cursor_on = false;
				repaintCell( cursor );
				cursor -= cursor % xsize;
			}
			else if ( c == '\n' ) {
				cursor_on = false;
				repaintCell( cursor );
				cursor += xsize;
				fixCursor();
			}
			else if ( c == 7 ) {
				QApplication::beep();
			}
 			else if ( c == 8 ) {
				if ( cursor % xsize ) {
					cursor_on = false;
					repaintCell( cursor );
					cursor--;
					data[cursor].c = ' ';
					cellChanged( cursor );
				}
			}
			else if ( c == '\t' ) {
				cursor_on = false;
				repaintCell( cursor );
				int cellCount = 8 - cursor % xsize % 8;
				for ( int i = 0; i < cellCount; i++ ) {
					outp_at( cursor, ' ' );
					cellChanged( cursor++ );
				}
			}
			else if ( c == 12 ) {
				for ( int j = 0; j < xsize * ysize; j++ ) outp_at( j, ' ' );
				cursor = 0;
				selectionStart = selectionEnd = -1;
				repaint( false );
			}
			else {
//				printf( "c=%02x\n", data[cursor].c.unicode() );
				outp_at( cursor, translateChar( c ) );
				cursor_on = false;
				cellChanged( cursor );
//				if ( cursor % xsize < xsize - 1 ) {
					cursor++;
					fixCursor();
//				}
			}
		}
		break;
	  }
  }
}

void TermEmu::fixCursor() {
  int i, pos;

  if ( cursor < xsize * ysize ) return;

//  printf( "fixing cursor (%d;%d)\n", cursor%xsize, cursor/xsize );

  while ( cursor >= xsize * ysize ) {
  	sbb->append( data );
  	memmove( data, data+xsize, (xsize * ysize - xsize) * sizeof( *data ) );

	for ( i = 0; i < xsize; i++ ) {
	  	pos = xsize*ysize-xsize + i;
		outp_at( pos, ' ' );
	}
  	cursor -= xsize;

	if ( hasSelection ) {
		selectionStart -= xsize;
		selectionEnd   -= xsize;
	}
  }

  repaint( FALSE );

//  printf( "fixing done (%d;%d)\n", cursor%xsize, cursor/xsize );
}

void TermEmu::drawCursor() {
  cursor_on = !cursor_on;
  repaintCell( cursor );
}

void TermEmu::cellChanged( int cell ) {
  if ( cell >= selectionStart && cell <  selectionEnd ||
       cell >  selectionEnd   && cell <= selectionStart ) {

	int minPos = selectionStart, maxPos = selectionEnd;
	if ( maxPos > minPos ) {
		minPos = selectionEnd;
		maxPos = selectionStart;
	}

	selectionStart = selectionEnd = -1;
	
	repaint( 0, minPos / xsize * cellh,
		width(), (maxPos / xsize - minPos / xsize + 1) * cellh,
		false );
  }
  else
	repaintCell( cell );
}

void TermEmu::repaintCell( int cell ) {
  int x, y;

  if ( cell >= xsize * ysize ) return;

  y = cell / xsize;
  x = cell % xsize;
  repaint( x * cellw, y * cellh, cellw, cellh );
}

static void if_swap( bool cond, int& v1, int& v2 ) {
  if ( cond ) {
	int tmp = v1;
	v1 = v2;
	v2 = tmp;
  }
}

void TermEmu::paintEvent( QPaintEvent *e ) {
  QPainter p;
  int i, j, pos, posx, posy;
  Cell *dp;
  int ux0, ux1, uy0, uy1, xp;
  int curs_x, curs_y, fg, bg, fg2, bg2;
  QChar buf[100], *bp;

  QFontMetrics fm( fnt );

  QRect r = e->rect();

  ux0 = r.x() / cellw;
  if ( ux0 < 0 ) ux0 = 0;

  ux1 = (r.x() + r.width() + cellw-1) / cellw;
  if ( ux1 > xsize ) ux1 = xsize;

  uy0 = r.y() / cellh;
  if ( uy0 < 0 ) uy0 = 0;

  uy1 = (r.y() + r.height() + cellh-1) / cellh;
  if ( uy1 > ysize ) uy1 = ysize;

  std::function<bool(int,int,int)> inSelectedArea;

  if ( hasSelection && (selectionType == LinesSelection || selectionType == ShortLinesSelection) ) {
	int start = selectionStart, end = selectionEnd;
	if_swap( start > end, start, end );

	inSelectedArea = [start, end](int pos, int posx, int posy) { return pos >= start && pos <= end; };
  }
  else if ( hasSelection && selectionType == RectSelection ) {
	int x0 = selectionStart % xsize,
	    y0 = selectionStart / xsize,
	    x1 = selectionEnd % xsize,
	    y1 = selectionEnd / xsize;
	if ( x0 < 0 ) {
		y0--;
		x0 += xsize;
	}
	if ( x1 < 0 ) {
		y1--;
		x1 += xsize;
	}

	if_swap( x0 > x1, x0, x1 );
	if_swap( y0 > y1, y0, y1 );

	inSelectedArea = [x0, x1, y0, y1](int pos, int posx, int posy) {
		 return posx >= x0 && posx <=  x1 && posy >= y0 && posy <= y1;
	};
  }
  else
	inSelectedArea = [](int, int, int) { return false; };


  QPixmap pm( (ux1-ux0)*cellw, (uy1-uy0)*cellh );

  p.begin( &pm );
  p.setBackgroundMode( OpaqueMode );
  p.setFont( fnt );

  for ( i = uy0; i < uy1; i++ ) {
	pos = (i - scrollbackLines) * xsize + ux0;
	if ( i < scrollbackLines )
		dp = sbb->line( scrollbackLines - 1 - i ) + ux0;
	else
		dp = data + pos;

	for ( j = ux0; j < ux1; ) {
		xp = j;
		bp = buf;

		for (; j < ux1; j++ ) {
			fg2 = dp->fg;
			bg2 = dp->bg;
			if_swap( inSelectedArea(pos, j, i - scrollbackLines), fg2, bg2 );

			if ( j == xp ) {
				fg = fg2;
				bg = bg2;
			}
			else {
				if ( fg2 != fg || bg2 != bg ) break;
			}
	
			*bp++ = dp->c;
			pos++;
			dp++;
		}

		p.setPen( col_pal[fg] );
		p.setBackgroundColor( col_pal[bg] );
		p.drawText( (xp - ux0)*cellw, (i - uy0)*cellh + fm.ascent(),
				QString( buf, bp - buf ) );
	}
  }

  if ( cursor_on ) {
  	curs_x = cursor % xsize;
	curs_y = cursor / xsize + scrollbackLines;
	if ( curs_x >= ux0 && curs_x <= ux1 && curs_y >= uy0 && curs_y <= uy1 ) {
  		QBrush b2( col_pal[data[cursor].fg] );
		p.fillRect( (curs_x-ux0)*cellw, (curs_y-uy0)*cellh, cellw, cellh, b2 );
	}
  }
  
  p.end();
 
  p.begin( this );
  p.drawPixmap( ux0*cellw, uy0*cellh, pm );
  p.end();
}

void TermEmu::focusInEvent( QFocusEvent* ) {
}

void TermEmu::focusOutEvent( QFocusEvent* ) {
}

void TermEmu::keyPressEvent( QKeyEvent *e ) {
  if ( e->key() == Key_B && e->state() == AltButton ) {
  	if ( sl ) tcsendbreak( sl->fd(), 0 );
	return;
  }
  else if ( e->key() == Key_X && e->state() == AltButton ) {
  	qApp->quit();
	e->accept();
	return;
  }
  else if ( e->key() == Key_D && e->state() == AltButton ) {
  	int fd;

	if ( sl ) fd = sl->fd(); else fd = -1;

  	DialDirectory d( fd, this );
	DialEntry *connInfo0;

	connect( &d, SIGNAL(grabSerialLine()), SLOT(releaseLine()) );
	connect( &d, SIGNAL(releaseSerialLine()), SLOT(grabLine()) );
	d.exec();

	connInfo0 = d.connectedTo();
	if ( connInfo0 ) {
		connInfo = *connInfo0;
		connInfoValid = true;
		iemsiState = IEMSI_WAIT_IRQ;
		log( QString("connected to %1").arg( connInfo.name ) );
	}
	e->accept();
	return;
  }
  else if ( e->key() == Key_R && e->state() == AltButton ) {
	e->accept();
	rxFileXModem();
	return;
  }
  else if ( e->key() == Key_S && e->state() == AltButton ) {
	QSettings s;
	s.setPath( "vr5", "term" );

	e->accept();

	SelectProtoDlg d( this );
	switch ( xferProto ) {
	case XferXmodem:   d.xmodemRB  ->setChecked( true ); break;
	case XferXmodem1k: d.xmodem1kRB->setChecked( true ); break;
	case XferYmodem:   d.ymodemRB  ->setChecked( true ); break;
	default:           d.zmodemRB  ->setChecked( true ); break;
	}

	if ( d.exec() != QDialog::Accepted ) return;

	if ( d.xmodemRB->isChecked() ) xferProto = XferXmodem;
	else if ( d.xmodem1kRB->isChecked() ) xferProto = XferXmodem1k;
	else if ( d.ymodemRB  ->isChecked() ) xferProto = XferYmodem;
	else xferProto = XferZmodem;

	if ( xferProto == XferZmodem ) {
		QStringList lst =  QFileDialog::getOpenFileNames( s.readEntry("term/txFromPath"),
			QString::null, this );
		if ( lst.count() ) {
			QFileInfo fi( lst[0] );
			s.writeEntry("term/txFromPath", fi.absFilePath());
			txFilesZModem( lst );
		}
	}
	else {
		QString filename = QFileDialog::getOpenFileName( s.readEntry("term/txFromPath"),
			QString::null, this );
		if ( !filename.isEmpty() ) {
			QFileInfo fi( filename );
			s.writeEntry("term/txFromPath", fi.absFilePath());

			switch (xferProto) {
			case XferXmodem:   txFilesXModem  ( filename ); break;
			case XferXmodem1k: txFilesXModem1k( filename ); break;
			case XferYmodem:   txFilesYModem  ( filename ); break;
			}
		}
	}

	return;
  }
  else if ( e->key() == Key_P && e->state() == AltButton ) {
	int rc, id;
  	PortProfileSelector w( this );

	rc = w.exec();
	if ( rc == QDialog::Accepted ) {
		id = w.getSelectedProfileId();
		if ( id != -1 ) {
			statLine->disable();
			sl->reopen( profiles[id].device.latin1(),
				 profiles.speedCode(id), profiles[id].flow );
		}
	}
	e->accept();
	return;
  }
  else if ( e->key() == Key_C && e->state() == (ControlButton | AltButton) ) {
	cursor = 0;
	eraseScreenPart( 0, xsize * ysize );
	e->accept();
	return;
  }
  else if ( e->key() == Key_Insert && e->state() == ShiftButton ) {
	QClipboard *cb = QApplication::clipboard();

	QString s = cb->text( QClipboard::Selection );
	if ( s.isEmpty() ) return;

	QByteArray arr = codec? codec->fromUnicode( s ): s.local8Bit();
	write( sl->fd(), arr, arr.size() - 1 );

	e->accept();
	return;
  }
  else if ( e->key() == Key_E && e->state() == AltButton ) {
	localEcho = !localEcho;
	e->accept();
	return;
  }
  else if ( e->key() == Key_I && e->state() == AltButton ) {
	if ( !initstr.isEmpty() ) {
		write( sl->fd(), initstr, initstr.size()-1 );
		write( sl->fd(), "\r", 1 );
	}
	e->accept();
	return;
  }
  else if ( e->key() == Key_H && e->state() == AltButton ) {
	if ( !hangupInProgress ) {
		int flags;
		ioctl( sl->fd(), TIOCMGET, &flags );
		ioctl( sl->fd(), TIOCMSET, flags & ~TIOCM_DTR );

		hangupInProgress = true;
		QTimer::singleShot( 1000, this, SLOT(finishHangup()) );
	}
	e->accept();
	return;
  }
  else if ( e->key() == Key_PageUp && e->state() == ShiftButton ) {
	int lines = scrollbackLines;
	scrollbackLines += ysize / 2;
	if ( scrollbackLines > sbb->lines )
		scrollbackLines = sbb->lines;
	if ( scrollbackLines != lines ) repaint( false );
	e->accept();
	return;
  }
  else if ( e->key() == Key_PageDown && e->state() == ShiftButton ) {
	int lines = scrollbackLines;
	scrollbackLines -= ysize / 2;
	if ( scrollbackLines < 0 ) scrollbackLines = 0;
	if ( scrollbackLines != lines ) repaint( false );
	e->accept();
	return;
  }
  if ( !sl ) {
  	e->accept();
  	return;
  }
  
  if ( e->key() == Key_Up ) write( sl->fd(), "\033[A", 3 );
  else if ( e->key() == Key_Down   ) write( sl->fd(), "\033[B", 3 );
  else if ( e->key() == Key_Right  ) write( sl->fd(), "\033[C", 3 );
  else if ( e->key() == Key_Left   ) write( sl->fd(), "\033[D", 3 );
  else if ( e->key() == Key_Home   ) write( sl->fd(), "\033[H", 3 );
  else if ( e->key() == Key_Insert ) write( sl->fd(), "\033[L", 3 );
  else if ( e->key() == Key_F1     ) write( sl->fd(), "\033[[A", 4 );
  else if ( e->key() == Key_F2     ) write( sl->fd(), "\033[[B", 4 );
  else if ( e->key() == Key_F3     ) write( sl->fd(), "\033[[C", 4 );
  else if ( e->key() == Key_F4     ) write( sl->fd(), "\033[[D", 4 );
  else if ( e->key() == Key_F5     ) write( sl->fd(), "\033[[E", 4 );
  else if ( e->key() == Key_F6     ) write( sl->fd(), "\033[17~", 5 );
  else if ( e->key() == Key_F7     ) write( sl->fd(), "\033[18~", 5 );
  else if ( e->key() == Key_F8     ) write( sl->fd(), "\033[19~", 5 );
  else if ( e->key() == Key_F9     ) write( sl->fd(), "\033[20~", 5 );
  else if ( e->key() == Key_F10    ) write( sl->fd(), "\033[21~", 5 );
  else if ( !e->text().isEmpty() ) {
	QByteArray arr = codec? codec->fromUnicode( e->text() ): e->text().local8Bit();
	write( sl->fd(), arr, arr.size() - 1 );
	if ( localEcho ) addStr( (unsigned char*)arr.data(), arr.size() - 1 );
  }

  e->accept();
}

void TermEmu::finishHangup() {
  int flags;
  ioctl( sl->fd(), TIOCMGET, &flags );
  ioctl( sl->fd(), TIOCMSET, flags | TIOCM_DTR );
  hangupInProgress = false;
}

void TermEmu::mousePressEvent( QMouseEvent *e ) {
  if ( e->button() == LeftButton ) {
	int startPrev = selectionStart,
	    endPrev   = selectionEnd;
	if ( endPrev < startPrev ) {
		int tmp = startPrev;
		startPrev = endPrev;
		endPrev = tmp;
	}

//        selectionType = (e->state() & ControlButton)? RectSelection: LinesSelection;
        selectionType = (e->state() & ControlButton)? RectSelection: ShortLinesSelection;

	selectionStart = selectionEnd  = (e->y() / cellh - scrollbackLines) * xsize + e->x() / cellw;
	hasSelection = true;
	selecting = true;

	int paintPos = (startPrev + scrollbackLines * xsize) / xsize * cellh,
            paintH   = ((endPrev - startPrev + xsize - 1) / xsize + 1) * cellh;

	if ( paintPos < 0 ) paintPos = 0;

	if ( paintH > 0 )
		repaint( 0, paintPos, width(), paintH, false );
  }
  else if ( e->button() == MidButton ) {
	QClipboard *cb = QApplication::clipboard();

	QString s = cb->text( QClipboard::Selection );
	if ( s.isEmpty() ) return;

	QByteArray arr = codec? codec->fromUnicode( s ): s.local8Bit();
  	write( sl->fd(), arr, arr.size() - 1 );
  }
}

void TermEmu::mouseReleaseEvent( QMouseEvent *e ) {
  if ( e->button() != LeftButton ) return;

  selecting = false;

  int minPos = selectionStart, maxPos = selectionEnd;
  if_swap( maxPos < minPos, minPos, maxPos );

  QString s;

  if ( selectionType == LinesSelection ) {
	for ( int i = minPos; i <= maxPos; i++ )
		if ( i < 0 )
			s += sbb->line((-i) / xsize)[xsize - (-i) % xsize].c;
		else
			s += data[i].c;
  }
  else if ( selectionType == ShortLinesSelection ) {
	int x0 = minPos % xsize,
            y0 = minPos / xsize,
            x1 = maxPos % xsize,
            y1 = maxPos / xsize;
	if ( x0 < 0 ) {
		y0--;
		x0 += xsize;
	}
	if ( x1 < 0 ) {
		y1--;
		x1 += xsize;
	}

	for ( int i = y0; i <= y1; i++ ) {
		Cell *line;
		if ( i < 0 )
			line = sbb->line(-i-1);
		else
			line = &data[i * xsize];

		int x_beg = (i == y0)? x0: 0;

		int x_end = (i == y1)? x1: xsize-1;
		while ( x_end >= x_beg && line[x_end].c == ' ' ) x_end--;

		for ( int j = x_beg; j <= x_end; j++ )
			s += line[j].c;

		if ( x_end != xsize-1 && i != y1 ) s += '\n';
	}
	
  }
  else if ( selectionType == RectSelection ) {
	int x0 = minPos % xsize,
            y0 = minPos / xsize,
            x1 = maxPos % xsize,
            y1 = maxPos / xsize;
	if ( x0 < 0 ) {
		y0--;
		x0 += xsize;
	}
	if ( x1 < 0 ) {
		y1--;
		x1 += xsize;
	}

	for ( int i = y0; i <= y1; i++ ) {
		Cell *line;
		if ( i < 0 )
			line = sbb->line(-i-1);
		else
			line = &data[i * xsize];

		for ( int j = x0; j <= x1; j++ )
			s += line[j].c;

		s += '\n';
	}
  }

  QClipboard *cb = QApplication::clipboard();
  cb->setText( s, QClipboard::Clipboard );
  cb->setText( s, QClipboard::Selection );

  hasSelection = selectionStart != selectionEnd;
}

void TermEmu::mouseMoveEvent( QMouseEvent *e ) {
  if ( !selecting ) return;

  int minPos = selectionStart,
      maxPos = selectionEnd;
  if ( maxPos < minPos ) {
	int tmp = maxPos;
	maxPos  = minPos;
	minPos  = tmp;
  }

  int evX = e->x(), evY = e->y();
  if ( evX < 0 ) evX = 0;
  if ( evX >= width() ) evX = width() - 1;

  if ( evY < 0 ) evY = 0;
  if ( evY >= height() ) evY = height() - 1;

  selectionEnd = (evY / cellh - scrollbackLines) * xsize + evX / cellw;

  if ( selectionEnd < minPos ) minPos = selectionEnd;
  else if ( selectionEnd > maxPos ) maxPos = selectionEnd;

  repaint( 0, (minPos + scrollbackLines * xsize) / xsize * cellh,
           width(), ((maxPos - minPos + xsize - 1) / xsize + 1) * cellh,
	   false );
}

void TermEmu::processCdDrop() {
  if ( !connInfoValid ) return;

  log( "disconnected" );
  connInfoValid = false;
}

void TermEmu::processFileRxStart( XferEvent *e ) {
  log( QString("starting to download %1").arg( e->fname ) );
}

void TermEmu::processFileRxResume( XferEvent *e ) {
  log( QString("resuming from %1 bytes").arg( e->bytes ) );
}

void TermEmu::processFileRxDone() {
  log( "finished downloading" );
}

void TermEmu::processFileTxStart( XferEvent *e ) {
  log( QString("starting to upload %1").arg( e->fname ) );
}

void TermEmu::processFileTxResume( XferEvent *e ) {
  log( QString("resuming from %1 bytes").arg( e->bytes ) );
}

void TermEmu::processFileTxDone() {
  log( "finished uploading" );
}

void TermEmu::log( const QString& msg0 ) {
  const char *mon_str[] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
  QString msg;
  QCString msgStr;

  QFile f( ConfigManager::getFilename("connect.log") );
  if ( !f.open( IO_ReadWrite | IO_Append ) ) return;

  QDateTime dt = QDateTime::currentDateTime();
  QString date = QString("yyyy-%1-dd hh:mm:ss ").arg( mon_str[dt.date().month()-1] );
  
  msg = dt.toString( date ) + msg0 + "\n";

  msgStr = msg.local8Bit();
  f.writeBlock( msgStr, msgStr.length() );
}

void TermEmu::saveSessionData( const QString& fnameTempl ) {
  QFile chrFile( fnameTempl + ".chr" );
  if ( !chrFile.open( IO_ReadWrite | IO_Truncate ) ) return;

  uint16_t *chrData = new uint16_t[xsize*ysize];
  for ( int i = 0; i < xsize*ysize; i++ )
  	chrData[i] = data[i].c.unicode();

  chrFile.writeBlock( (char*)chrData, xsize * ysize * sizeof(uint16_t) );
  chrFile.close();

  delete[] chrData;

  QFile attrFile( fnameTempl + ".attr" );
  if ( !attrFile.open( IO_ReadWrite | IO_Truncate ) ) return;

  uint8_t cursorData[2] = { (uint8_t)(cursor / xsize), (uint8_t)(cursor % xsize) };

  uint8_t *attrData = new uint8_t[xsize*ysize*2];
  for ( int i = 0; i < xsize*ysize; i++ ) {
  	attrData[i*2]   = data[i].fg;
  	attrData[i*2+1] = data[i].bg;
  }

  attrFile.writeBlock( (char*)cursorData, sizeof(cursorData)               );
  attrFile.writeBlock( (char*)attrData,   xsize * ysize * sizeof(uint16_t) );
  attrFile.close();

  delete[] attrData;
}

void TermEmu::restoreSessionData( const QString& fnameTempl ) {
  QFile chrFile( fnameTempl + ".chr" );
  if ( !chrFile.open( IO_ReadOnly ) ) return;

  uint16_t *chrData = new uint16_t[xsize*ysize];
  if ( chrFile.readBlock( (char*)chrData, xsize*ysize*sizeof(uint16_t) ) < xsize*ysize*sizeof(uint16_t) ) {
  	delete[] chrData;
  	return;
  }

  chrFile.close();

  for ( int i = 0; i < xsize*ysize; i++ )
  	data[i].c = chrData[i];

  delete[] chrData;

  QFile attrFile( fnameTempl + ".attr" );
  if ( !attrFile.open( IO_ReadOnly ) ) return;

  uint8_t cursorData[2];
  uint8_t *attrData = new uint8_t[xsize*ysize*2];

  if ( attrFile.readBlock( (char*)cursorData, sizeof(cursorData) ) < sizeof(cursorData) ) {
  	delete[] attrData;
  	return;
  }
  if ( attrFile.readBlock( (char*)attrData, xsize*ysize*2 ) < xsize*ysize*2  ) {
  	delete[] attrData;
  	return;
  }

  attrFile.close();

  cursor = cursorData[0] * xsize + cursorData[1];

  for ( int i = 0; i < xsize*ysize; i++ ) {
  	data[i].fg = attrData[i*2];
  	data[i].bg = attrData[i*2+1];
  }

  delete[] attrData;
}

class TerminalMainWindow: public QWidget {
public:
  TerminalMainWindow();
  ~TerminalMainWindow();
protected:
  TermEmu *term;
  SerialLine *line;
  QString sessionId;

  QFont selectFont();
  QString getSessionId();
  void keyPressEvent( QKeyEvent *e );
};

static void print_help() {
  printf( "Command line parameters:\n"
          "--font-family name  Use font from family \"name\" for terminal window\n"
          "--font-size sz      Use font with size \"sz\" for terminal window\n"
          "--session id        Use session \"id\" for saving/restoring session data\n"
          "--initstring str    Set init string for modems\n"
          "--dialing-type t    Set dialing type ('t' tone, 'p' pulse, '' default) for dialer\n" );
}

TerminalMainWindow::TerminalMainWindow(): line(0) {
  profiles.load();

  {
	QSettings s;
	s.setPath( "vr5", "term" );
	for ( int i = 0; i < qApp->argc(); i++ )
		if ( !strcmp( qApp->argv()[i], "--initstr" ) && ++i < qApp->argc() ) {
			s.writeEntry( "term/initstr", qApp->argv()[i] );
			break;
		}
		else if ( !strcmp( qApp->argv()[i], "--dialing-type" ) && ++i < qApp->argc() ) {
			s.writeEntry( "term/dialingType", qApp->argv()[i] );
			break;
		}
		else if ( !strcmp( qApp->argv()[i], "--help" ) ) {
			print_help();
			exit(0);
		}
  }

  term = new TermEmu( this, selectFont() );
  term->move( 0, 0 );
  term->show();

  StatusLine *statLine = new StatusLine( this );
  term->setStatusLine( statLine );

  statLine->adjustSize();
  statLine->setGeometry( 0, term->height(), term->width(), statLine->height() );

  setFixedSize( term->width(), term->height() + statLine->height() );
  show();

  line = new SerialLine;

  QString portDevice = "/dev/ttyLT";
  int portSpeed = B115200,
	portFlow = PortProfile::FLOW_NONE;

  if ( currentPortProfile >= 0 && currentPortProfile < profiles.size() ) {
	portDevice = profiles[currentPortProfile].device;
	portSpeed = profiles.speedCode(currentPortProfile);
	portFlow = profiles[currentPortProfile].flow;
  }

  if ( line->open( portDevice, portSpeed, portFlow ) == -1 ) {
  	perror( "line open" );
	statLine->disable();
	line->reopen();
  }

  connect( statLine, SIGNAL(cdDropped()), term, SLOT(processCdDrop()) );
  statLine->attachLine( line );
  connect( line, SIGNAL(lineReopened(SerialLine*)),
	 statLine, SLOT(attachLine(SerialLine*)) );

  sessionId = getSessionId();
  term->restoreSessionData( ConfigManager::getVarPath( sessionId ) );

  term->attachLine( line );

  new XModemDBusNodeAdapter( term );
}

TerminalMainWindow::~TerminalMainWindow() {
  profiles.save();
  delete line;

  term->saveSessionData( ConfigManager::getVarPath( sessionId ) );
}

QFont TerminalMainWindow::selectFont() {
  QSettings s;
  s.setPath( "vr5", "term" );

  QString fontFamily = s.readEntry( "term/fontFamily", "Courier" );
  int     fontSize   = s.readEntry( "term/fontSize",   "-1"      ).toInt();

  for ( int i = 0; i < qApp->argc(); i++ )
  	if ( !strcmp( qApp->argv()[i], "--font-family" ) && ++i < qApp->argc() ) {
		fontFamily = qApp->argv()[i];
		s.writeEntry( "term/fontFamily", fontFamily );
	}
  	else if ( !strcmp( qApp->argv()[i], "--font-size" ) && ++i < qApp->argc() ) {
		fontSize = atoi( qApp->argv()[i] );
		s.writeEntry( "term/fontSize", fontSize );
	}

  if ( fontSize <= 0 ) fontSize = 12;

  QFont fnt = QFont( fontFamily, fontSize );
  fnt.setFixedPitch( true );

  return fnt;
}

QString TerminalMainWindow::getSessionId() {
  QString sessionId = "session";
  for ( int i = 0; i < qApp->argc(); i++ )
  	if ( !strcmp( qApp->argv()[i], "--session" ) && ++i < qApp->argc() ) {
		sessionId += QString("-") + qApp->argv()[i];
		break;
	}
  return sessionId;
}

void TerminalMainWindow::keyPressEvent( QKeyEvent *e ) {
  QWidget::keyPressEvent( e );
}

QString ConfigManager::config;
QString ConfigManager::var;

ConfigManager::ConfigManager() {
  QString home = getenv( "HOME" );
  if ( !home.isEmpty() ) {
  	config = home + "/.config/term/";
	var    = home + "/.term/";

	mkpath( config );
	mkpath( var );
	mkpath( var + "files" );
  }
}

void ConfigManager::mkpath( const QString& d ) {
  QStringList lst = QStringList::split( '/', QDir::cleanDirPath( d ) );
  QDir qd;
  QString p;
  for ( int i = 0; i < lst.size(); i++ ) {
  	p += "/" + lst[i];
  	qd.mkdir( p );
  }
}

QString ConfigManager::getFilename( const QString& confName ) {
  return config + confName;
}

QString ConfigManager::getVarPath( const QString& fname ) {
  return var + fname;
}

int main( int argc, char **argv ) {
  QApplication qa( argc, argv );

  ConfigManager conf;

  TerminalMainWindow w;
  qa.setMainWidget( &w );

  qa.exec();

  return 0;
}
