/*	
 * Copyright 2001-2010 Georges Menie (www.menie.org)
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* this code needs standard functions memcpy() and memset()
   and input/output functions _inbyte() and _outbyte().

   the prototypes of the input/output functions are:
     int _inbyte(unsigned short timeout); // msec timeout
     void _outbyte(int c);

 */

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include "xmodemio.h"
#include "crc/crctab.h"

#define SOH  0x01
#define STX  0x02
#define EOT  0x04
#define ACK  0x06
#define NAK  0x15
#define CAN  0x18
#define CTRLZ 0x1A

#define DLY_1S 1000
#define MAXRETRANS 25

static void nop() {}

void (*XModem_feFileInfo)( const char *fname, int bytes ) = nop;
void (*XModem_feFileData)( int bytes )                    = nop;
void (*XModem_feFileDone)()                               = nop;

char XModem_cancelled = 0;
  
unsigned short crc16_ccitt(const void *buf, int len)
{
	register int counter;
	register unsigned short crc = 0;
	for( counter = 0; counter < len; counter++)
		updcrc( *(char*)buf++, crc );
	return crc;
}

static int check(int crc, const unsigned char *buf, int sz)
{
	if (crc) {
		unsigned short crc = crc16_ccitt(buf, sz);
		unsigned short tcrc = (buf[sz]<<8)+buf[sz+1];
		if (crc == tcrc)
			return 1;
	}
	else {
		int i;
		unsigned char cks = 0;
		for (i = 0; i < sz; ++i) {
			cks += buf[i];
		}
		if (cks == buf[sz])
		return 1;
	}

	return 0;
}

static void flushinput(void)
{
	char buf[512];
	while (_inbytes(buf, 512, DLY_1S>>2) >= 0)
		;
}

static char xfer_cancelled() {
  if ( XModem_cancelled ) {
	flushinput();
	_outbyte(CAN);
	_outbyte(CAN);
	_outbyte(CAN);
  }

  return XModem_cancelled;
}

int xmodemReceive( const char *fname )
{
	unsigned char xbuff[1030]; /* 1024 for XModem 1k + 3 head chars + 2 crc + nul */
	int bufsz, crc = 0;
	unsigned char trychar = 'C';
	unsigned char packetno = 1;
	int i, c, len = 0;
	int retry, retrans = MAXRETRANS;
	FILE *f;
	int rc = 0;

	f = fopen( fname, "wb" );
	if ( !f ) return -6;

	XModem_feFileInfo( fname, -1 );

	for(;;) {
		for( retry = 0; retry < 16; ++retry) {
			if ( xfer_cancelled() ) goto done;

			if (trychar) _outbyte(trychar);
			if ((c = _inbyte((DLY_1S)<<1)) >= 0) {
				switch (c) {
				case SOH:
					bufsz = 128;
					goto start_recv;
				case STX:
					bufsz = 1024;
					goto start_recv;
				case EOT:
					flushinput();
					_outbyte(ACK);
					XModem_feFileDone();
					rc = len; /* normal end */
					goto done;
				case CAN:
					if ((c = _inbyte(DLY_1S)) == CAN) {
						flushinput();
						_outbyte(ACK);
						rc = -1; /* canceled by remote */
						goto done;
					}
					break;
				default:
					flushinput();
					break;
				}
			}
		}
		if (trychar == 'C') { trychar = NAK; continue; }
		_outbyte(CAN);
		_outbyte(CAN);
		_outbyte(CAN);
		flushinput();
		rc = -2; /* sync error */
		goto done;

	start_recv:
		if (trychar == 'C') crc = 1;
		trychar = 0;
		xbuff[0] = c;
		if (_inbytes(xbuff+1, bufsz+(crc?1:0)+3, DLY_1S) < 0) goto reject;

		if (xbuff[1] == (unsigned char)(~xbuff[2]) && 
		   (xbuff[1] == packetno || xbuff[1] == (unsigned char)packetno-1) &&
			check(crc, &xbuff[3], bufsz)) {
			if (xbuff[1] == packetno)	{
				fwrite( &xbuff[3], 1, bufsz, f );
				len += bufsz;
				XModem_feFileData( len );
				++packetno;
				retrans = MAXRETRANS+1;
			}
			if (--retrans <= 0) {
				_outbyte(CAN);
				_outbyte(CAN);
				_outbyte(CAN);
				flushinput();
				rc = -3; /* too many retry error */
				goto done;
			}
			_outbyte(ACK);
			continue;
		}

	reject:
		_outbyte(NAK);
		flushinput();
	}

done:
  fclose( f );
  return rc;
}

int xmodemTransmit(const char *fname, char vers_1k )
{
	unsigned char xbuff[1030]; /* 1024 for XModem 1k + 3 head chars + 2 crc + nul */
	int bufsz, crc = -1;
	unsigned char packetno = 1;
	int i, c, len = 0;
	int retry;
	FILE *f;
	int rc = 0;
	char eof = 0;
	struct stat stat_buf;
	off_t fsize = -1;

	f = fopen( fname, "rb" );
	if ( !f ) return -6;

	if ( stat( fname, &stat_buf ) != -1 )
		fsize = stat_buf.st_size;

	XModem_feFileInfo( fname, fsize );

	for(;;) {
		for( retry = 0; retry < 16; ++retry) {
			if ( xfer_cancelled() ) goto done;

			if ((c = _inbyte((DLY_1S)<<1)) >= 0) {
				switch (c) {
				case 'C':
					crc = 1;
					goto start_trans;
				case NAK:
					crc = 0;
					goto start_trans;
				case CAN:
					if ((c = _inbyte(DLY_1S)) == CAN) {
						_outbyte(ACK);
						flushinput();
						rc = -1; /* canceled by remote */
						goto done;
					}
					break;
				default:
					break;
				}
			}
		}
		_outbyte(CAN);
		_outbyte(CAN);
		_outbyte(CAN);
		flushinput();
		rc = -2; /* no sync */
		goto done;

		for(;;) {
		start_trans:
			if ( xfer_cancelled() ) goto done;

			if ( vers_1k ) {
				xbuff[0] = STX; bufsz = 1024;
			}
			else {
				xbuff[0] = SOH; bufsz = 128;
			}
			xbuff[1] = packetno;
			xbuff[2] = ~packetno;
                        if ( !eof ) {
				memset (&xbuff[3], 0, bufsz);
				c = fread( &xbuff[3], 1, bufsz, f );
				if (c == -1) c = 0;
				if (c < bufsz) {
					xbuff[3+c] = CTRLZ;
					eof = 1;
				}
				if (crc) {
					unsigned short ccrc = crc16_ccitt(&xbuff[3], bufsz);
					xbuff[bufsz+3] = (ccrc>>8) & 0xFF;
					xbuff[bufsz+4] = ccrc & 0xFF;
				}
				else {
					unsigned char ccks = 0;
					for (i = 3; i < bufsz+3; ++i) {
						ccks += xbuff[i];
					}
					xbuff[bufsz+3] = ccks;
				}
				for (retry = 0; retry < MAXRETRANS; ++retry) {
					_outbytes(xbuff, bufsz+4+(crc?1:0));
					if ((c = _inbyte((DLY_1S) << 1)) >= 0 ) {
						switch (c) {
						case ACK:
							++packetno;
							len += bufsz;
							XModem_feFileData( len );
							goto start_trans;
						case CAN:
							if ((c = _inbyte(DLY_1S)) == CAN) {
								_discardoutput();
								_outbyte(ACK);
								flushinput();
								rc = -1; /* canceled by remote */
								goto done;
							}
							break;
						case NAK:
							_discardoutput();
							_inbyte(DLY_1S >> 1); /* delay to let receiver flush its input */
						default:
							break;
						}
					}
				}
				_outbyte(CAN);
				_outbyte(CAN);
				_outbyte(CAN);
				flushinput();
				rc = -4; /* xmit error */
				goto done;
			}
			else {
				for (retry = 0; retry < 10; ++retry) {
					_outbyte(EOT);
					if ((c = _inbyte((DLY_1S)<<1)) == ACK) break;
				}
//				flushinput();

				XModem_feFileDone();

				rc = (c == ACK)?len:-5;
				goto done;
			}
		}
	}

done:
	fclose( f );
	return rc;
}

int xmodemTransmit1k(const char *fname) {
  xmodemTransmit( fname, 1 );
}

int xmodemTransmit128(const char *fname) {
  xmodemTransmit( fname, 0 );
}

#ifdef TEST_XMODEM_RECEIVE
int main(void)
{
	int st;

	printf ("Send data using the xmodem protocol from your terminal emulator now...\n");
	/* the following should be changed for your environment:
	   0x30000 is the download address,
	   65536 is the maximum size to be written at this address
	 */
	st = xmodemReceive((char *)0x30000, 65536);
	if (st < 0) {
		printf ("Xmodem receive error: status: %d\n", st);
	}
	else  {
		printf ("Xmodem successfully received %d bytes\n", st);
	}

	return 0;
}
#endif
#ifdef TEST_XMODEM_SEND
int main(void)
{
	int st;

	printf ("Prepare your terminal emulator to receive data now...\n");
	/* the following should be changed for your environment:
	   0x30000 is the download address,
	   12000 is the maximum size to be send from this address
	 */
	st = xmodemTransmit((char *)0x30000, 12000);
	if (st < 0) {
		printf ("Xmodem transmit error: status: %d\n", st);
	}
	else  {
		printf ("Xmodem successfully transmitted %d bytes\n", st);
	}

	return 0;
}
#endif
