#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <utime.h>
#include <sys/time.h>
#include <sys/stat.h>

#include "crc/crctab.h"
#include "zmodem.h"

#define XON  ('q' & 0x1f)
#define XOFF ('s' & 0x1f)

#define ZPAD   '*'
#define ZDLE   24
#define ZDLEE  (ZDLE ^ 0x40)
#define ZBIN   'A'
#define ZHEX   'B'
#define ZBIN32 'C'

#define ZRQINIT 0
#define ZRINIT  1
#define ZACK    3
#define ZFILE   4
#define ZSKIP   5
#define ZNAK	6
#define ZABORT  7
#define ZFIN    8
#define ZRPOS   9
#define ZDATA	10
#define ZEOF    11
#define ZCRC    13

#define CANFC32 0x20

#define ZCRCE 'h' // frame ends, header packet follows
#define ZCRCG 'i' // frame continues nonstop
#define ZCRCQ 'j' // frame continues, ZACK expected
#define ZCRCW 'k' // ZACK expected, end of frame

#define ZF1_ZMCRC 2

#define ZM_HEADER   0
#define ZM_NOHEADER 1
#define ZM_CANCEL   2
#define ZM_DATA     3
#define ZM_OVERFLOW 4
#define ZM_BADCRC   5
#define ZM_LSKIP    6
#define ZM_LCANCEL  7
#define ZM_ERROR    8
#define ZM_TIMEOUT  9
#define ZM_HANGUP   10

bool tx_reverse_crc = true;

ZModem::ZModem( int read_fd0, int write_fd0 ):
	read_fd( read_fd0 ), write_fd( write_fd0 ), zdle_count(0),
	reverse_crc(false), crc32Mode(false), foundHdr32(false) {
  dlpath = strdup( "files" );
  pipe( cmdPipe );
}

ZModem::~ZModem() {
  free( dlpath );
  close( cmdPipe[0] );
  close( cmdPipe[1] );
}

int ZModem::readFromLine( char *buf, int len, int timeout ) {
  timeval tv, tv_start, tv_cur;
  fd_set rfd;
  int rc, maxfd, elap_time;
  char c, idx;

  if ( read_fd > cmdPipe[0] ) maxfd = read_fd; else maxfd = cmdPipe[0];

  idx = 0;
  gettimeofday( &tv_start, 0 );
  for (;;) {
	if ( timeout == -1 )
		tv.tv_sec = 10;
	else {
		gettimeofday( &tv_cur, 0 );

		elap_time = tv_cur.tv_sec - tv_start.tv_sec;
		if ( elap_time >= timeout ) return ZM_TIMEOUT;

		tv.tv_sec = timeout - elap_time;
	}
	
	tv.tv_usec = 0;

  	FD_ZERO( &rfd );
	FD_SET( read_fd, &rfd );
	FD_SET( cmdPipe[0], &rfd );

  	rc = select( maxfd+1, &rfd, 0, 0, &tv );
	if ( !rc && timeout != -1 ) return ZM_TIMEOUT;
	if ( rc == -1 ) return ZM_ERROR;

	if ( FD_ISSET( cmdPipe[0], &rfd ) ) {
		read( cmdPipe[0], &c, 1 );
		if ( c == 1 ) return ZM_LSKIP;
		if ( c == 2 ) return ZM_LCANCEL;
	}
	if ( FD_ISSET( read_fd, &rfd ) ) {
		rc = read( read_fd, buf+idx, 1 );
		if ( !rc ) {
			puts( "zm hangup" );
			return ZM_HANGUP;
		}
		idx++;
		if ( idx >= len ) break;
	}
  }

  return 0;
}

void ZModem::putHexHdr( int type, char *data ) {
  char buf[7];
  int i, crc;

  buf[0] = ZPAD;
  buf[1] = ZPAD;
  buf[2] = ZDLE;
  buf[3] = ZHEX;
  write( write_fd, buf, 4 );

  sprintf( buf, "%02x", type );
  write( write_fd, buf, 2 );
  crc = updcrc( type & 0xff, 0 ) & 0xffff;
 
  for ( i = 0; i < 4; i++ ) {
  	sprintf( buf, "%02x", data[i] & 0xff );
	write( write_fd, buf, 2 );
	crc = updcrc( data[i] & 0xff, crc ) & 0xffff;
  }
  crc = updcrc( 0, crc ) & 0xffff;
  crc = updcrc( 0, crc ) & 0xffff;

  sprintf( buf, "%02x%02x\r%c", (crc >> 8) & 0xff, crc & 0xff, XON );
  write( write_fd, buf, 6 );
}

void ZModem::putHexHdr( int type ) {
  char flags[4];

  memset( flags, 0, 4 );
  putHexHdr( type, flags );
}

void ZModem::zdleWrite( char *data, int len ) {
  unsigned char c;

  for ( int i = 0; i < len; i++ ) {
	c = data[i];
  	if ( c == ZDLE || c == XON || c == XOFF ||
	     c == (XON | 0x80) || c == (XOFF | 0x80) ) {
		c = ZDLE;
		write( write_fd, &c, 1 );
		c = data[i] ^ 0x40;
		write( write_fd, &c, 1 );
	}
	else 
  		write( write_fd, &c, 1 );
  }
}

void ZModem::putBin16Hdr( int type, char *data ) {
  char buf[4];
  int i, crc;

  buf[0] = ZPAD;
  buf[1] = ZDLE;
  buf[2] = ZBIN;
  buf[3] = type;
  write( write_fd, buf, 4 );

  zdleWrite( data, 4 );

  crc = updcrc( type & 0xff, 0 ) & 0xffff;
  for ( i = 0; i < 4; i++ ) crc = updcrc( data[i] & 0xff, crc ) & 0xffff;
  crc = updcrc( 0, crc ) & 0xffff;
  crc = updcrc( 0, crc ) & 0xffff;

  buf[0] = crc >> 8;
  buf[1] = crc & 0xff;
  zdleWrite( buf, 2 );
}

void ZModem::putBin32Hdr( int type, char *data ) {
  char buf[4];
  int i, crc;

  buf[0] = ZPAD;
  buf[1] = ZDLE;
  buf[2] = ZBIN32;
  buf[3] = type;
  write( write_fd, buf, 4 );

  zdleWrite( data, 4 );

  crc = updcrc32( type, 0xffffffff );
  for ( i = 0; i < 4; i++ ) crc = updcrc32( data[i], crc );
  crc = ~crc;

  zdleWrite( (char*)&crc, 4 );
}

void ZModem::putBinHdr( int type, char *data ) {
  if ( crc32Mode ) putBin32Hdr( type, data ); else putBin16Hdr( type, data );
}

void ZModem::putBinHdr( int type ) {
  char flags[4];

  memset( flags, 0, 4 );
  putBinHdr( type, flags );
}

void ZModem::putData16( char *data, int len, int endtype ) {
  int i, crc;
  char buf[2];

  zdleWrite( data, len );

  crc = 0;
  for ( i = 0; i < len; i++ ) crc = updcrc( data[i] & 0xff, crc ) & 0xffff;
  crc = updcrc( endtype, crc ) & 0xffff;
  crc = updcrc( 0, crc ) & 0xffff;
  crc = updcrc( 0, crc ) & 0xffff;
  
  buf[0] = ZDLE;
  buf[1] = endtype;
  write( write_fd, buf, 2 );

  buf[0] = crc >> 8;
  buf[1] = crc & 0xff;
  zdleWrite( buf, 2 );
}

void ZModem::putData32( char *data, int len, int endtype ) {
  int i, crc;
  char buf[2];

  zdleWrite( data, len );

  crc = 0xffffffff;
  for ( i = 0; i < len; i++ ) crc = updcrc32( data[i], crc );
  crc = ~updcrc32( endtype, crc );

  buf[0] = ZDLE;
  buf[1] = endtype;
  write( write_fd, buf, 2 );
  zdleWrite( (char*)&crc, 4 );
}

void ZModem::putData( char *data, int len, int endtype ) {
  if ( crc32Mode ) putData32( data, len, endtype );
  else putData16( data, len, endtype );
}

int ZModem::getXByte( char *s ) {
  char buf[3];
  int rc;

  buf[0] = s[0];
  buf[1] = s[1];
  buf[2] = 0;

  return strtol( buf, 0, 16 );
}

bool debug_gethdr = false;

int ZModem::getHdr( int& type, char *data, bool wait, int timeout ) {
  char c, buf[14];
  const char *xdig = "0123456789abcdef";
  int i, rc, state, buf_ptr, crc, recvd_crc, elap_time;
  timeval tv_start, tv_cur;
  bool got_zdle;

  state = 0;
  gettimeofday( &tv_start, 0 );
  for (;;) {
  	if ( !wait && state == 0 && !dataAvail() ) return ZM_NOHEADER;

	if ( timeout == -1 ) 
		rc = readFromLine( &c, 1 );
	else {
		gettimeofday( &tv_cur, 0 );
		elap_time = tv_cur.tv_sec - tv_start.tv_sec;
		if ( elap_time >= timeout ) return ZM_TIMEOUT;

  		rc = readFromLine( &c, 1, timeout - elap_time );
	}
	if ( rc ) return rc;

	if ( debug_gethdr ) printf( "getHdr: c=%02x, state=%d\n", c, state );
	if ( isZCancel( c ) ) return ZM_CANCEL;

	switch ( state ) {
	case 0:
		if ( c == ZPAD ) state = 1;
		break;
	case 1:
		if ( c == ZDLE ) state = 2;
		else if ( c != ZPAD ) state = 0;
		break;
	case 2:
		if ( c == ZBIN ) {
			puts( "gethdr binhdr" );
			got_zdle = false;
			state = 3;
			buf_ptr = 0;
		}
		else if ( c == ZBIN32 ) {
			puts( "gethdr bin32hdr" );
			foundHdr32 = true;
			got_zdle = false;
			state = 4;
			buf_ptr = 0;
		}
		else if ( c == ZHEX ) {
			puts( "gethdr hexhdr" );
			state = 5;
			buf_ptr = 0;
		}
		else state = 0;

		break;
	case 3:
		if ( c == ZDLE ) {
			got_zdle = true;
			break;
		}
		if ( got_zdle ) c ^= 0x40;
		got_zdle = false;
	
		buf[buf_ptr++] = c;
		if ( buf_ptr == 7 ) {		// type, payload[4], crc[2]
			type = buf[0];
			memcpy( data, buf+1, 4 );

			crc = updcrc( type & 0xff, 0 ) & 0xffff;
			for ( i = 0; i < 4; i++ )
				crc = updcrc( data[i] & 0xff, crc ) & 0xffff;
		
			if ( reverse_crc ) {
				crc = updcrc( buf[6] & 0xff, crc ) & 0xffff;
				crc = updcrc( buf[5] & 0xff, crc ) & 0xffff;
			}
			else {
				crc = updcrc( buf[5] & 0xff, crc ) & 0xffff;
				crc = updcrc( buf[6] & 0xff, crc ) & 0xffff;
			}
		
			if ( crc != 0 ) {
				puts( "gethdr1 badcrc" );
				return ZM_BADCRC;
			}
		
			return ZM_HEADER;
		}
		break;
	case 4:
		if ( c == ZDLE ) {
			got_zdle = true;
			break;
		}
		if ( got_zdle ) c ^= 0x40;
		got_zdle = false;
	
		buf[buf_ptr++] = c;
		if ( buf_ptr == 9 ) {		// type, payload[4], crc[4]
			type = buf[0];
			memcpy( data, buf+1, 4 );

			crc = updcrc32( type, 0xffffffff );
			for ( i = 1; i < 9; i++ )
				crc = updcrc32( buf[i], crc );
		
			if ( crc != 0xdebb20e3 ) {
				puts( "gethdr2 badcrc" );
				return ZM_BADCRC;
			}
		
			return ZM_HEADER;
		}
		break;
	case 5:
		buf[buf_ptr++] = c;
		if ( !strchr( xdig, c ) ) {
			state = 0;
			break;
		}
		if ( buf_ptr == 14 ) {		// type[2], payload[8], crc[4]
			type = getXByte( buf );

			crc = updcrc( type & 0xff, 0 ) & 0xffff;
			if ( debug_gethdr ) printf( "gethdr c=%d, crc=%04x\n", type, crc );
			for ( i = 0; i < 4; i++ ) {
				data[i] = getXByte( &buf[(i+1)*2] );
				crc = updcrc( data[i] & 0xff, crc ) & 0xffff;
				if ( debug_gethdr ) printf( "gethdr c=%d, crc=%04x\n", data[i], crc );
			}
			crc = updcrc( getXByte( &buf[10] ), crc ) & 0xffff;
			crc = updcrc( getXByte( &buf[12] ), crc ) & 0xffff;
	
			if ( crc != 0 ) {
				if ( debug_gethdr )
					puts( "gethdr3 badcrc" );
				return ZM_BADCRC;
			}
			
			return ZM_HEADER;
		}
		break;
	}
  }
}

bool ZModem::isZCancel( char c ) {
  if ( c == ZDLE ) {
	zdle_count++;
	return zdle_count >= 5;
  }

  zdle_count = 0;
  return false;
}

int ZModem::getData( char *data, int& len, int& endtype, int pos ) {
  int rc, clen, state, crcidx, crc, crc_recvd;
  char c, crcbuf[4], posbuf[4];

  clen = 0;
  state = 0;
  if ( crc32Mode ) crc = 0xffffffff; else crc = 0;
  memset( crcbuf, 0, 4 );
  for (;;) {
	rc = readFromLine( &c, 1 );
	if ( rc ) return rc;

//	printf( "rxdata c=%d, state=%d\n", c, state );
	switch ( state ) {
	case 0:
		if ( isZCancel( c ) ) return ZM_CANCEL;

		if ( c == ZDLE )
			state = 1;
		else {
			if ( clen >= len ) return ZM_OVERFLOW;
			data[clen++] = c;
		
			if ( crc32Mode ) crc = updcrc32( c, crc );
			else crc = updcrc( c & 0xff, crc ) & 0xffff;
	
//			printf( "rxdata c=%02x, crc=%08x\n", c, crc );
		}
		
		break;
	case 1:
		if ( isZCancel( c ) ) return ZM_CANCEL;

		if ( c == ZCRCE || c == ZCRCG || c == ZCRCQ || c == ZCRCW ) {
			len = clen;
			endtype = c;

			if ( crc32Mode ) crc = updcrc32( c, crc );
			else crc = updcrc( c & 0xff, crc ) & 0xffff;
		
			state = 2;
			crcidx = 0;
			break;
		}
	
		c ^= 0x40;

		if ( clen >= len ) return ZM_OVERFLOW;
		data[clen++] = c;
	
		if ( crc32Mode ) crc = updcrc32( c, crc );
		else crc = updcrc( c & 0xff, crc ) & 0xffff;
//		printf( "rxdata c=%02x, crc=%08x\n", c, crc );
	
		state = 0;
		break;
	case 2:
		if ( isZCancel( c ) ) return ZM_CANCEL;

		if ( c == ZDLE )
			state = 3;
		else {
			crcbuf[crcidx++] = c;
			
			if ( crc32Mode ) {
				if ( crcidx >= 4 ) goto finish_processing;
			}
			else
				if ( crcidx >= 2 ) goto finish_processing;
		}
		break;
	case 3:
		if ( isZCancel( c ) ) return ZM_CANCEL;

		crcbuf[crcidx++] = c ^ 0x40;
	
		if ( crc32Mode ) {
			if ( crcidx >= 4 ) goto finish_processing;
		}
		else
			if ( crcidx >= 2 ) goto finish_processing;
		
		state = 2;
	}
  }

finish_processing:
  if ( crc32Mode ) {
  	int i;

  	for ( i = 0; i < 4; i++ )
		crc = updcrc32( crcbuf[i], crc );
	
	if ( crc != 0xdebb20e3 ) {
		puts( "rxdata32 badcrc" );
		return ZM_BADCRC;
	}
  }
  else {
	if ( reverse_crc ) {
		crc = updcrc( crcbuf[1] & 0xff, crc ) & 0xffff;
		crc = updcrc( crcbuf[0] & 0xff, crc ) & 0xffff;
	}
	else {
		crc = updcrc( crcbuf[0] & 0xff, crc ) & 0xffff;
		crc = updcrc( crcbuf[1] & 0xff, crc ) & 0xffff;
	}

	if ( crc != 0 ) {
	  	puts( "rxdata badcrc" );
	  	return ZM_BADCRC;
	}
  }

  if ( endtype == ZCRCQ || endtype == ZCRCW ) {
  	pos += len;
	*(int*)posbuf = pos;
  	putHexHdr( ZACK, posbuf );
  }

  return ZM_DATA;
}

bool ZModem::dataAvail() {
  int rc;
  fd_set rfd;
  timeval tv;

  FD_ZERO( &rfd );
  FD_SET( read_fd, &rfd );
  
  tv.tv_sec = 0;
  tv.tv_usec = 1;

  rc = select( read_fd+1, &rfd, 0, 0, &tv );
  if ( rc < 1 ) return false;

  return FD_ISSET( read_fd, &rfd );
}

int ZModem::putZFile( int& f, char *fname ) {
  char buf[1024], *s, *p;
  struct stat st;

  f = open( fname, O_RDONLY );
  if ( f == -1 ) return 1;

  putBinHdr( ZFILE );

  p = buf;
  s = strrchr( fname, '/' );
  if ( !s ) s = fname; else s++;
  strcpy( p, s );
  p += strlen( p ) + 1;
			
  fstat( f, &st );
  sprintf( p, "%d %o %o", st.st_size, st.st_mtime, st.st_mode );
  p+= strlen( p );

  putData( buf, p - buf, ZCRCW );

  feFileInfo( s, st.st_size );

  return 0;
}

void ZModem::sendFiles( char **flist ) {
  char buf[1024], flags[4], *p, *s;
  int f, type, state, offs, rc, fnum;
  struct stat st;
  bool all_done, wait_hdr;

  zdle_count = 0;
  putHexHdr( ZRQINIT );
  all_done = false;

  reverse_crc = tx_reverse_crc;

  for (;;) {
	rc = getHdr( type, flags, true ); 
	printf( "txfile rc=%d, type=%d state=1\n", rc, type );
	if ( rc == ZM_CANCEL || rc == ZM_LCANCEL ||
		rc == ZM_ERROR || rc == ZM_HANGUP ) return;
	if ( rc == ZM_HEADER && type == ZRINIT ) break;
  }
  if ( flags[3] & CANFC32 ) crc32Mode = true;

  fnum = 0;
  printf( "txfile fname=%s\n", flist[fnum] );
  while ( flist[fnum] ) {
	if ( putZFile( f, flist[fnum] ) ) {
		fnum++;
		continue;
	}
	
	for (;;) {
		rc = getHdr( type, flags, true );
		printf( "txfile rc=%d, type=%d state=2\n", rc, type );
		if ( rc == ZM_CANCEL || rc == ZM_ERROR ||
			rc == ZM_HANGUP ) return;
		if ( rc == ZM_LCANCEL ) {
			cancelSession();
			return;
		}
		if ( rc == ZM_HEADER && type == ZSKIP )
			goto send_next_file;
		if ( rc == ZM_HEADER && type == ZRPOS ) {
			offs = *(int*)flags;
			if ( offs ) feFileResume( offs );
			goto process_zrpos;
		}
	}

	for (;;) {
		rc = getHdr( type, flags, false );
		printf( "txfile rc=%d, type=%d state=3\n", rc, type );
		if ( rc == ZM_CANCEL || rc == ZM_ERROR ||
			rc == ZM_HANGUP ) return;
		if ( rc == ZM_LCANCEL ) {
			cancelSession();
			return;
		}
		if ( rc == ZM_HEADER && type == ZRPOS ) {
process_zrpos:
			offs = *(int*)flags;
			lseek( f, offs, SEEK_SET );
			putBinHdr( ZDATA, (char*)&offs );
		}
		else if ( rc == ZM_HEADER && type == ZSKIP )
			goto send_next_file;
		
		while ( !dataAvail() ) {
			feFileData( offs );

			rc = read( f, buf, 1024 );
			if ( !rc ) {
				putData( 0, 0, ZCRCE );
				putBinHdr( ZEOF, (char*)&offs );
				goto all_data_sent;
			}
			else {
				putData( buf, rc, ZCRCG );
				offs += rc;
			}
		}
	}
all_data_sent:
	for (;;) {
		rc = getHdr( type, flags, true );
		printf( "txfile rc=%d, type=%d state=4\n", rc, type );
		if ( rc == ZM_CANCEL || rc == ZM_ERROR ||
			rc == ZM_HANGUP ) return;
		if ( rc == ZM_LCANCEL ) {
			cancelSession();
			return;
		}
		if ( rc == ZM_HEADER && type == ZRINIT ) break;
		if ( rc == ZM_HEADER && type == ZRPOS )
			goto process_zrpos;
	}
	if ( flags[3] & CANFC32 ) crc32Mode = true; else crc32Mode = false;

send_next_file:
	close( f );
	feFileDone();

	fnum++;
  }

  putBinHdr( ZFIN );
  for (;;) {
  	getHdr( type, flags, true );
	if ( rc == ZM_CANCEL || rc == ZM_ERROR || rc == ZM_HANGUP ) return;
	if ( rc == ZM_LCANCEL ) {
		cancelSession();
		return;
	}
	if ( rc == ZM_HEADER && type == ZFIN ) break;
  }

  write( write_fd, "OO", 2 );
}

void ZModem::cancelSession() {
  char cancel_seq[] = {
  	ZDLE, ZDLE, ZDLE, ZDLE, ZDLE, ZDLE, ZDLE, ZDLE,
	8, 8, 8, 8, 8, 8, 8, 8 };
  int rc, type;
  char flags[4];

  for (;;) {
  	putHexHdr( ZABORT );
  	rc = getHdr( type, flags, true, 10 );
	if ( rc == ZM_CANCEL || rc == ZM_ERROR || rc == ZM_HANGUP ) break;
	if ( rc == ZM_TIMEOUT ) {
		write( write_fd, cancel_seq, sizeof cancel_seq );
		break;
	}
	if ( rc == ZM_HEADER && type == ZFIN ) break;
  }
}

int ZModem::skipFile() {
  int rc, type;
  char flags[4];
 
  putHexHdr( ZSKIP );
  for (;;) {
  	rc = getHdr( type, flags, true );
	printf( "rxskip rc=%d, type=%d\n", rc, type );
	if ( rc == ZM_CANCEL || rc == ZM_ERROR || rc == ZM_HANGUP ) return 1;
	if ( rc == ZM_LCANCEL ) {
		cancelSession();
		return 2;
	}
	if ( rc == ZM_HEADER && type == ZFIN ) {
		putHexHdr( ZFIN );
		return 1;
	}
	if ( rc == ZM_HEADER && type == ZRQINIT ) break;
  }

  return 0;
}

void createPath( char *path ) {
  char *s, *p;
  int rc;

  s = strdup( path );
  p = s;
  while ( p ) {
  	p = strchr( p, '/' );
	if ( p ) *p = 0;
  	rc = mkdir( s, 0755 );
	if ( rc == -1 ) break;
	if ( p ) *p = '/';
  }
  free( s );
}

int ZModem::getFileCrc( int fd, int len ) {
  char buf[1024];
  int rc, pos, blen, crc, i;

  crc = 0xffffffff;
  while ( pos < len ) {
  	if ( pos + 1024 > len ) blen = len - pos; else blen = 1024;
	rc = read( fd, buf, blen );
	for ( i = 0; i < rc; i++ )
		crc = updcrc32( buf[i], crc );
	pos += rc;
  }
  return ~crc;
}

// returns 0 if remote system has file with the same crc
//         1 in case of error, hangup, cancel
//	   2 if remote system has file with different crc
//					(resuming is not possible )
int ZModem::zrCheckCrc( int f, int len ) {
  int crc, type, rc, rcrc;

  crc = getFileCrc( f, len );
  putHexHdr( ZCRC );

  rc = getHdr( type, (char*)&rcrc, true, 1 );
  printf( "zrCheckCrc rc=%d, type = %d\n", rc, type );
  if ( rc == ZM_CANCEL || rc == ZM_LCANCEL ||
	rc == ZM_ERROR || rc == ZM_HANGUP ) return 1;
  if ( rc == ZM_TIMEOUT || rc == ZM_HEADER && type != ZCRC ) return 2;
  if ( crc != rcrc ) {
  	printf( "zrCheckCrc crc=%08x, rcrc=%08x\n", crc, rcrc );
	return 2;
  }

  return 0;
}

// return 0 if everithing is ok
//        1 in case of error, hangup, cancel
//        2 if skipping file
int ZModem::startZReceive( char *fname, int& f, int& pos, int& mtime ) {
  char buf[1025], *s, *p;
  int len, type, filesize, rc;
  struct stat st;

  len = sizeof(buf) - 1;
  rc = getData( buf, len, type, 0 );
  printf( "rxfile zdata rc=%d, type='%c'\n", rc, type );
  if ( rc == ZM_CANCEL || rc == ZM_ERROR || rc == ZM_HANGUP ) return 1;
  if ( rc == ZM_OVERFLOW || rc == ZM_BADCRC ) {
	putHexHdr( ZNAK );
	return 2;
  }
  if ( rc == ZM_LCANCEL ) {
  	cancelSession();
	return 1;
  }

  s = strrchr( buf, '/' );
  if ( !s ) s = buf; else s++;

  sprintf( fname, "%s/%s", dlpath, s );

  f = open( fname, O_CREAT|O_RDWR, 0644 );
  if ( f == -1 && errno == ENOENT ) {
  	createPath( dlpath );
  	f = open( fname, O_CREAT|O_RDWR, 0644 );
  }
  if ( f == -1 ) {
	putHexHdr( ZSKIP );
	return 2;
  }

  rc = fstat( f, &st );
  if ( rc == -1 ) {
	putHexHdr( ZSKIP );
	return 2;
  }

  buf[len] = 0;

  p = buf + strlen( buf ) + 1;
  filesize = 0;
  sscanf( p, "%d %o", &filesize, &mtime );
  feFileInfo( s, filesize );

  pos = st.st_size;

  if ( pos != 0 && remoteHasZMCRC ) {
  	rc = zrCheckCrc( f, pos );
	if ( rc == 1 ) return 1;
	if ( rc == 2 ) pos = 0;
  }

  if ( pos != 0 ) feFileResume( pos );

  putHexHdr( ZRPOS, (char*)&pos );
  lseek( f, pos, SEEK_SET );

  return 0;
}

int ZModem::recvFileData( int f, int &pos ) {
  char buf[1024];
  int len, rc, type;

  for (;;) {
	len = sizeof buf;
	rc = getData( buf, len, type, pos );
//	printf( "rxfile zdata rc=%d, len=%d, type='%c'\n", rc, len, type );
	if ( rc == ZM_CANCEL || rc == ZM_ERROR || rc == ZM_HANGUP ) return 1;
	if ( rc == ZM_LCANCEL ) {
		cancelSession();
		return 1;
	}
	if ( rc == ZM_LSKIP ) {
		rc = skipFile();
		if ( rc == 1 ) return 1;
		return 2;
	}
	if ( rc == ZM_OVERFLOW || rc == ZM_BADCRC ) {
		putHexHdr( ZRPOS, (char*)&pos );
		break;
	}
		
	write( f, buf, len );
	pos += len;

	feFileData( pos );

	if ( type == ZCRCE || type == ZCRCW ) break; 
  }

  return 0;
}

void ZModem::zrCloseFile( int f, char *fname, int mtime ) {
  utimbuf utbuf;

  close( f );
  utbuf.actime = mtime;
  utbuf.modtime = mtime;
  utime( fname, &utbuf );
}

void ZModem::receiveFiles( bool skip_state0 ) {
  int f, rc, type, len, pos, filesize, mtime;
  char flags[4], lflags[4], fname[1024];

  memset( lflags, 0, 4 );
  lflags[3] |= CANFC32;

  reverse_crc = false;
  foundHdr32 = false;
  for (;;) {
init_session:
	crc32Mode = false;
	if ( !skip_state0 )
		putHexHdr( ZRINIT, lflags );
	skip_state0 = false;

wait_zfile:
	for (;;) {
		putHexHdr( ZRINIT, lflags );
		crc32Mode = false;
	
		rc = getHdr( type, flags, true );
		printf( "rxfile rc=%d, type=%d\n", rc, type );
		if ( rc == ZM_CANCEL || rc == ZM_ERROR ||
			rc == ZM_HANGUP ) goto zrx_done;
		if ( rc == ZM_LCANCEL ) {
			cancelSession();
			return;
		}
		if ( foundHdr32 ) crc32Mode = true;
	
		if ( rc == ZM_HEADER ) {
			if ( type == ZFILE ) break;
			else if ( type == ZFIN ) {
				putHexHdr( ZFIN );
				goto zrx_done;
			}
			else if ( type == ZRINIT )
				goto zrx_done;
		}	

		putHexHdr( ZRINIT, lflags );
		crc32Mode = false;
	}

	printf( "zfile flags=0x%02x,0x%02x,0x%02x,0x%02x\n",
		flags[0], flags[1], flags[2], flags[3] );

	if ( flags[2] & ZF1_ZMCRC ) remoteHasZMCRC = true;
	else remoteHasZMCRC = false;

	rc = startZReceive( fname, f, pos, mtime );
	if ( rc == 1 ) break;
	if ( rc == 2 ) goto wait_zfile;

	for (;;) {
		rc = getHdr( type, flags, true );
		printf( "rxfile rc=%d, type=%d\n", rc, type );
		if ( rc == ZM_CANCEL || rc == ZM_ERROR ||
			rc == ZM_HANGUP ) goto zrx_done;
		if ( rc == ZM_LCANCEL ) {
			cancelSession();
			zrCloseFile( f, fname, mtime );
			return;
		}
		if ( rc == ZM_LSKIP ) {
			zrCloseFile( f, fname, mtime );
			rc = skipFile();
			if ( rc == 1 ) return;
			goto init_session;
		}

		if ( rc != ZM_HEADER ) continue;
		
		if ( type == ZRQINIT ) {
			zrCloseFile( f, fname, mtime );
			goto init_session;
		}
		else if ( type == ZEOF ) {
			if ( pos == *(int*)flags ) break;
			putHexHdr( ZRPOS, (char*)&pos );
		}
		else if ( type == ZDATA ) {
			if ( pos != *(int*)flags ) {
				putHexHdr( ZRPOS, (char*)&pos );
				continue;
			}
			rc = recvFileData( f, pos );
			if ( rc == 1 ) {
				zrCloseFile( f, fname, mtime );
				goto zrx_done;
			}
			if ( rc == 2 ) {
				zrCloseFile( f, fname, mtime );
				goto init_session;
			}
		}
	}

	zrCloseFile( f, fname, mtime );
	feFileDone();
  }
zrx_done:;
}

#ifdef MAIN
int pipe_test() {
  int rx[2], tx[2];
  pid_t child;

  pipe( rx );
  pipe( tx );
  child = fork();
  if ( !child ) {
  	close( rx[0] );
	close( tx[1] );

	ZModem zm( tx[0], rx[1] );
	zm.receiveFiles( false );
	puts( "done receiving" );
	return 0;
  }
  close( rx[1] );
  close( tx[0] );

  char *file_list[] = { "/temp/RPMS-1", "/temp/RPMS-2", 0 };
  
  ZModem zm( rx[0], tx[1] );
  zm.sendFiles( file_list );
  puts( "done sending" );
  return 0;
}

#include <termios.h>

void tty_test( bool rx ) {
  int fd = open( "/dev/ttyS1", O_RDWR );
  if ( fd == -1 ) {
  	perror( "open" );
  	return;
  }
  struct termios ts;
  tcgetattr( fd, &ts );
  cfmakeraw( &ts );
  cfsetospeed( &ts, B9600 );
  cfsetispeed( &ts, B9600 );
  tcsetattr( fd, TCSANOW, &ts );

  ZModem zm( fd, fd );
  if ( rx ) zm.receiveFiles( false );
  else {
	char *file_list[] = { "/tmp/korolew_enru.dwa", 0 };
	zm.sendFiles( file_list );
  }
}

int main() {
//  pipe_test();
//  debug_gethdr = true;
//  tty_test( false );
  tty_test( true );
}
#endif
